/*
This file - CMB82.c - serves as the main program in the C++/Fortran system of 
files that the Delphi client (CMB82.exe) for EPA-CMB8.2 calls at run time.
CMB82.s receives computational results from CMB82main.for and CMB82supp.for 
and formats for the "spreadsheet" output.  Several run-time warning messages 
also come from CMB82.c.                                                -TCoulter

setSPcode
setPRcode
setADcode  set and write
setPROcode set and write

Of note (Norm Robinson):

Atotal(i,j) holds the profiles: i=i, ... ,nSPtot; j=1, numprorc (# of stored profiles)
The ordering in i is that of ELcode.  i=1 is for total - all entries are 1.0

PROcode holds the column in atotal for each source/size profile.

The profile direct access file holds PROcode and atotal info.  All of this can
be replaced with a single structure.

VARIABLES:
sizeindex - C++     - index of active size fraction 0,1,2,3
sizendx   - Fortran - index of active size fraction 1,2,3,4
*/

// DLL interface to the Fortran code:

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <time.h>
#include <sys\stat.h>
#include "WKShead.h"

#pragma warning(disable: 4018) // Signed/Unsigned mismatch.

/* ==========================================================================*/
// Constants.

#define LENSTRINGS 256
#define LENERRBUF   80
#define LenDefSel   19

// indices begin with 0
#define indexCMB8ADfile    6 // CMB8.2 ambient data index in Control File
#define indexCMB8profile   7 // CMB8.2 source profile index in Control File
#define indexCMB8SPselect  4 // CMB8.2 element selection index in Control File
#define indexCMB8PRselect  3 // CMB8.2 source selection index in Control File
#define indexCMB8ADselect  5 // CMB8.2 ambient data selection index in Control File

#define ERROR_MEMORY_ALLOC      -1
#define ERROR_FILE_OPEN         -2
#define ERROR_SPECIES_SET       -3
#define ERROR_SOURCES_SET       -4
//#define ERROR_CMBSVD_1          -5 TC shut down 11/21/04; no longer needed
#define ERROR_CMBSVD_2          -6
//#define ERROR_CMBSVD_3          -7    *** this ERROR is never set anywhere!
#define ERROR_CMBSVD_4          -8
#define ERROR_CMBSVD_5          -9
#define ERROR_CMBSVD_6         -10
#define ERROR_FILE_READ        -11
#define ERROR_MISC             -12
#define ERROR_FILE_CLOSE       -13
#define ERROR_FILE_WRITE       -14
#define ERROR_AMBIENT_SET      -15
#define ERROR_PROFILE_SET      -16
//#define ERROR_READ_SPECIES_SEL -17
//#define ERROR_READ_SOURCES_SEL -18
//#define ERROR_READ_AMBDATA_SEL -19
//#define ERROR_READ_AMBDATA     -20
#define ERROR_READ_PROFILE     -21
#define ERROR_MEMORY_FREE      -22

#define lenSPcode  24
#define lenSPcode1  6 // length of first token - code
#define lenSPcode2  8 // length of second token - name

#define lenPRcode  24
#define lenPRcode1  6 // length of first token
#define lenPRcode2  8 // length of second token

#define lenADcode  42 // ADCode is a 42-char string amenable to an
                      // internal read in CMB82supp.for's ???????? 
#define lenADcode1 12 // site ID
#define lenADcode2  8 // date
#define lenADcode3  2 // duration
#define lenADcode4  2 // start hour
#define lenADcode5  6 // size fraction
#define lenADcode6  6 // record number // record number in (binary) file AMBdirect.dat

// PROcode is a char. string that stores the source profile records to be used; 
                       // it is internally read	in CMB82supp.for's s.LoadATOT
#define lenPROcode  26 // expaned from 25 to accommodate 6-char size fraction
#define lenPROcode1  6 // source profile number
#define lenPROcode2  8 // source profile name
#define lenPROcode3  6 // size fraction; changed from 5 10/2004 -TC
#define lenPROcode4  6 // record number in (binary) file PROdirect.dat

#define LENTOKEN   25
#define MAXTOKENS 512

#define MAXSIZES        4
#define INDEXSIZENAME  28 // index  of size fraction name in ADcode (=> col. 29)
#define LENSIZENAME     6 // length of size fraction name in ADcode
#define MAXFILELENGTH 128

/* ==========================================================================*/
// Enumerations:
// NB: The following parameters MUST correspond exactly with those listed in
// the fRun declarations (@65) in CMB82main.for, as well as 
// the Internal Constants (@150) listed in Delphi's CMB82Internals.pas:

#define DO_CMB82init       1
#define DO_SendCodes       2
#define DO_SetDisplayDecPl 3
#define DO_SetBatchFlag    4
#define DO_SetDichotFlag   5 // TC added 10-24-04
#define DO_CMB             6
#define DO_SendFracEst     7
#define DO_SendInfo        8
#define DO_SendSNGVAL      9
#define DO_PData          10
#define DO_SScont         11
#define DO_mPIN           12
#define DO_SFitOut        13
#define DO_CSV2TXT        14
#define DO_DBF2TXT        15
#define DO_WKS2TXT        16
#define DO_CAR2TXT        17
#define DO_CMB_Terminate  18

/* ==========================================================================*/
// Safety macros:

#define Fclose(X) if (X) { Fclose(X); X = 0; }
#define FREE(X) { free(X); X = NULL; }

/* ==========================================================================*/
// Data structures:
// xBASE structures:
// First 32 bytes of xBASE header:
typedef struct
{
    char version;     //   3 = dBASE or FoxPro - no memo field
                      // 245 = FoxPro with memo field
                      // 131 = dBASE III or FoxPlus with memo field
                      // 139 = dBASE IV with memo field
    char yy;
    char mm;
    char dd;
    BYTE numrecs0;    // LSB (Least Significant Bytes)
    BYTE numrecs1;
    BYTE numrecs2;
    BYTE numrecs3;    // MSB (Most  Significant Bytes)
    BYTE headlen0;    // LSB (Least Significant Bytes)
    BYTE headlen1;    // MSB (Most  Significant Bytes)
    BYTE reclen0;     // LSB (Least Significant Bytes)
    BYTE reclen1;     // MSB (Most  Significant Bytes)
    BYTE _unused[20];
} XBASE_HEADER;

// xBASE field description:
typedef struct
{
    char name[12]; // field name
    char type;     // field type:
                   // 'C' - character
                   // 'D' - date
                   // 'F' - floating point(binary)
                   // 'L' - logical
                   // 'N' - floating point(bcd)
    int  len;      // field width
    int  decpl;    // decimal count
    int  offset;   // offset into data buffer
    // position 0 for delete flag
} XBASE_FIELDS;

// Fortran string.
typedef struct
{
    char *addr;
    DWORD len;
} FSTRING;

/* ========================================================================== */
// Globals:

static char    messagetitle[] = "CMB82.c";

static char    CMB8exedir[256];
static char    ProDirName[256];
static char    AmbDirName[256];
static char    SumDirName[256];

static FSTRING ftn_char_SPcode;
static FSTRING ftn_char_SPcode_user;
static FSTRING ftn_char_PRcode;
static FSTRING ftn_char_PRcode_user;
static FSTRING ftn_char_ADcode;
static FSTRING ftn_char_PROcode;
static FSTRING ftn_char_ambname;
static FSTRING ftn_char_AmbDirName;
static FSTRING ftn_char_proname;
static FSTRING ftn_char_ProDirName;
static FSTRING ftn_char_SumDirName;
static FSTRING ftn_char_sizes;
static FSTRING ftn_char_outbuf;
static FSTRING ftn_char_errbuf;
static FSTRING ftn_char_SFit;

static int     SFitLen;
static int     SFitRecLen;
static char    SFitName[128];

static char   *cfp0;
static char   *cfp1;
static char   *cfp2;
static char   *cfp3;
static char   *cfp4;
static char   *cfp5;
static char   *cfp6;
static char   *cfp7;
static char   *cfp8;
static char   *cfp9;

static char    cbuf[LENSTRINGS];
static char    cbuf1[LENSTRINGS];
static char    cbuf2[LENSTRINGS];

// CMB8.2 parameters and storage arrays:
static char(*talkarray)[LENSTRINGS];
static char(*SPcode)[lenSPcode];
static char(*SPcode_user)[lenSPcode];
static char(*SPcodetmp)[lenSPcode];
static char(*PRcode)[lenPRcode];
static char(*PRcode_user)[lenPRcode];
static char(*PRcodetmp)[lenPRcode];
static char(*ADcode)[lenADcode];
static char(*ADcodetmp)[lenADcode];
static char(*SPdefsel)[LenDefSel];
static char(*PRdefsel)[LenDefSel];
static char(*PROcode)[lenPROcode];
static char(*tokens1)[LENTOKEN];
static char(*tokens2)[LENTOKEN];
static char(*tokens3)[LENTOKEN];

static char   *Cwork;
static char   *Cwork1;
static char   *outbuf;
static char   *SFitBuf;
static char   *errbuf;
static int    *iwork1;
static int    *iSPfit;
static int    *jPRfit;
static int    *inestim; // tags inestimable sources - 1 for inestimable
static float  *CMBwork;
static float  *work2;
static float  *atotal;
static float  *uatotal;
static float  *ambSP;
static float  *uambSP;
static float  *sngval;

static int     nSPtot;
static int     nPRtot;
static int     nADtot;
static int     nPROtot;
static int     nSPprofile;
static int     nSPambient; // number of elements in data files

static int     nSPfit;
static int     nPRfit;
static int     numCMBfit;

static int     dimoutbuf = 1;
static int     dimerrbuf = 1;

static char    sizearr[MAXSIZES + 1][LENSIZENAME];
static int     numsizes;

static int     dimCwork;
static FSTRING ftn_char_Cwork;
static int     dimiwork1;
static int     dimwork;
static int     dimCMBwork;

static int     ambrec;
static int     sizeindex;

static int     lenoutbuf;
static int     lenerrbuf;

static int     SFitbas;
static int     uSFitbas;

static int     sourcefitflag;
static int     ambrecflag;  // flag whether ambient records have been selected
static int     DichotFlagC; // Added 10/26/04; only needed here is pass via args.

static int     iDummy;
static int     Write2DT2status;

static FILE   *pFile;

static char    infile[MAXFILELENGTH];
static char    outfile[MAXFILELENGTH];

static char    winbuf[255];

// Defined for WKS output file.
static char    label[100] =
{
  '\017', '\000', '\000', '\000', '\000',
  '\000', '\000', '\000', '\000', '\'',
  '\000'
};

static char    integer[11] =
{
  '\015', '\000', '\007', '\000', '\004',
  '\000', '\000', '\000', '\000', '\000',
  '\000'
};

static char    real[17] =
{
  '\016', '\000', '\015', '\000', '\004',
  '\000', '\000', '\000', '\000', '\000',
  '\000'
};

/* ========================================================================== */
// Fortran declarations:

extern
int
__stdcall
INITIALIZE(
    int    *nSPtot,
    int    *nPRtot,
    int    *nADtot,
    int    *nPROtot,
    FSTRING ftn_char_PROcode,
    int    *numsizes,
    FSTRING ftn_char_sizes,
    FSTRING ftn_char_ambname,
    FSTRING ftn_char_AmbDirName,
    FSTRING ftn_char_proname,
    FSTRING ftn_char_ProDirName,
    FSTRING ftn_char_SumDirName,
    float  *atotal,
    float  *uatotal,
    int    *iDummy);

extern
int
__stdcall
FRUN(				 // 39 arguments
    int    *nSPtot,
    int    *nSPfit,
    FSTRING SPcode,
    FSTRING SPcode_user,
    int    *nPRtot,
    int    *nPRfit,
    FSTRING PRcode,
    FSTRING PRcode_user,
    int    *numCMBfit,
    int    *nPROtot,
    FSTRING PROcode,
    int    *nADtot,
    FSTRING ADcode,
    float  *ambSP,
    float  *uambSP,
    float  *atotal,
    float  *uatotal,
    float  *sngval,
    int    *iSPfit,
    int    *jPRfit,
    int    *inestim,
    int    *dimiwork1,
    int    *iwork1,
    int    *dimwork,
    float  *work,
    int    *sel,
    int    *action,
    int    *action1,
    FSTRING outbuf,
    int    *dimoutbuf,
    int    *lenoutbuf,
    FSTRING ftn_char_Cwork,
    FSTRING errbuf,
    int    *dimerrbuf,
    int    *lenerrbuf,
    int    *sfitbas,
    int    *usfitbas,
    int    *sourcefitflag,
    int    *DichotFlagC); // Added 10/26/04 for pass to CMB82main.for's s.PData

//  The following extern enables the call to CMB82supp.for's f.SetAmbRec:
extern
int
__stdcall
SETAMBREC(
    FSTRING ADcode);

extern
void
__stdcall
SFITHEAD(
    int    *action,
    int    *type,
    FSTRING PRcode,
    FSTRING ftp_char_sfit);

extern
void
__stdcall
SFITOUT(
    int    *action,
    int    *type,
    float  *sfit,
    float  *usfit,
    float  *atot,
    float  *uatot,
    int    *nSPtotPR,
    int    *iwork1,
    float  *ambSP,
    float  *uambSP,
    int    *jPRfit,
    FSTRING fnt_char_SPcode,
    FSTRING fnt_char_PRcode,
    FSTRING fnt_char_PROcode,
    float  *work,
    FSTRING fnt_char_sfit,
    int    *error);

// Common block for Fortran:

extern struct
{
    float CHIsqr;
    float Rsquare;
    float miss;
    float pcntMass;
    float fracestim;
    float maxSourceUnc;
    float minSourceProj;
    int   DisplayDecPl;
	float sumSCEs; // TC added 10/28/04
} COMMON_CMB;

/* ========================================================================== */
// Function declarations:

int tokcnt(char *cp);
int getmem(char *ptname, void **pt, size_t siz, unsigned int line, int action);
void __stdcall MessageBoxC_(FSTRING desc);
void message(char *title, char *message);
void message1(char *mess);
int SFitHead_c(int *type);
int SFitOut_c(int *type);
int SFitDBF_c();
int SFitWKS_c();
int csv2txt(char *infile, char *outfile);
int dbf2txt(char *infile, char *outfile);
int wks2txt(char *infile, char *outfile);
int car2txt(char *infile, char *outfile);
int AppendRec(int handle, char *dbfbuf);
int Create(int handle, XBASE_FIELDS *fields, int numfields, char **dbfbuf, int *reclen);
int loadDBinfo(int handle, XBASE_FIELDS **fields, int *p_numfields,unsigned int *p_numrecs, int *p_reclen, char **dbfbuf);
void NRdate(int *year, int *month, int *day);
int writeWKSheader();
int WKSlabel(unsigned short row, unsigned short col, char *cp);
int WKSreal(unsigned short row, unsigned short col, double value);
int StrCmpNR(char cp1[], char cp2[]);
int StuffTokArray(char *toks, int maxnum, int width, char *cp);
void SepToks(char *cp);
void Rjust(char *cp);
void Ljust(char *cp);
void Ljust1(char *cp, int len1); // left justify - keep trailing blanks
void FreeTempMemFP();
void ParseAmbient(char *toks, int width, char *ADcode);
int setSPcode();  // 1
int setPRcode();  // 2
int setADcode();  // 3
int setPROcode(); // 4
char *StrTokNR1(char *str, int *p_pos);

/* ========================================================================== */
// The only function exported by this DLL:

int
__stdcall
CMB82c(
    void *w0,
    void *w1,
    void *w2,
    void *w3,
    void *w4,
    void *w5,
    void *w6,
    void *w7,
    void *w8,
    void *w9,
    void *w10,
    void *w11,
    void *w12)
{
    static int   *Lfp1;
    static int   *Lfp2;
    static int   *Lfp3;
    static int   *Lfp4;
    static int   *Lfp5;
    static int   *Lfp6;
    static int   *Lfp7;
    static int   *Lfp8;
    static int   *Lfp9;
    static int   *Lfp10;
    static int   *Lfp11;
    static int   *Lfp12;
    static float *ffp0;
    static float *ffp1;
    static float *ffp2;
    static float *ffp3;
    static float *ffp4;
    static float *ffp5;
    static float *ffp6;
    static float *ffp7;
    static float *ffp8;
    static float *ffp9;
    static float *ffp10;
    static float *ffp11;
    static float *ffp12;
    static int    error;
    static int    sel;
    static size_t siz;
    static int    sizoutbuf;
    static int    action;
    static int    action1;
    static int    action2;
    static int    Ltmp;

    int command;
    int n;
    int len;

    command = *((int*)w0);
    switch (command)
    
    {

//	Step I
    case DO_CMB82init:
    /* delete direct access files to eliminate stale data */
    siz = 16;
    error = getmem("talkarray", (void **) &talkarray, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    siz = MAXTOKENS;
    error = getmem("tokens1", (void **) &tokens1, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    error = getmem("tokens2", (void **) &tokens2, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    error = getmem("tokens3", (void **) &tokens3, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

// These values passed in from Delphi's (CMB82Internals.pas) Task 3
    cfp3 = (char *) w4;  // Full Path for PR*.sel
    cfp4 = (char *) w5;  // Full Path for SP*.sel
    cfp5 = (char *) w6;  // Full Path for AD*.sel
    cfp6 = (char *) w7;  // Full Path for AD*.* (data input file)
    cfp7 = (char *) w8;  // Full Path for PR*.* (data input file)
    cfp8 = (char *) w9;  // string: CMBoutDB.txt
    Lfp10 = (int *) w10; // numCMBfit; set at end of CMB82main.for's s.CMBSVD (TC)
    iDummy = *Lfp10;
    if (iDummy == 0)
    {
      remove(SumDirName);
    }
	cfp9 = (char *) w11; // Full Path for executable directory

    strncpy((char *) talkarray[3], cfp3, LENSTRINGS);
    strncpy((char *) talkarray[4], cfp4, LENSTRINGS);
    strncpy((char *) talkarray[5], cfp5, LENSTRINGS);
    strncpy((char *) talkarray[6], cfp6, LENSTRINGS);
    strncpy((char *) talkarray[7], cfp7, LENSTRINGS);
    strncpy((char *) CMB8exedir, cfp9, 255);

                // Set path and name for direct access data files:
                len = strlen(CMB8exedir);
                strcpy(ProDirName, CMB8exedir);
                strcpy(ProDirName+len, "PROdirect.dat");

                strcpy(AmbDirName, CMB8exedir);
                strcpy(AmbDirName+len, "AMBdirect.dat");

                strcpy(SumDirName, CMB8exedir);
                strcpy(SumDirName+len, "SumDirect.dat");

    error = setSPcode();
        if (error)
        {
          return error;
        }

    error = setPRcode();
        if (error)
        {
          return error;
        }

    error = setADcode();  // Creates/populates AMBdirect.dat from *.sel file
        if (error)
        {
          return error;
        }

    error = setPROcode(); // Creates/populates PROdirect.dat from *.sel file
        if (error)
        {
          return error;
        }

//  Dichot. sample INVERT module (removed 10/2004 -TC):

    ftn_char_SPcode.addr = (char *) SPcode;
    ftn_char_SPcode.len = lenSPcode;

    siz = nSPtot+1;
    error = getmem("SPcode_user", (void **) &SPcode_user, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    ftn_char_SPcode_user.addr = (char *) SPcode_user;
    ftn_char_SPcode_user.len = lenSPcode;

    ftn_char_PRcode.addr = (char *) PRcode;
    ftn_char_PRcode.len = lenPRcode;

    siz = nPRtot+1;
    error = getmem("PRcode_user", (void **) &PRcode_user, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    ftn_char_PRcode_user.addr = (char *) PRcode_user;
    ftn_char_PRcode_user.len = lenPRcode;

    ftn_char_ADcode.addr = (char *) ADcode;
    ftn_char_ADcode.len = lenADcode;

    ftn_char_PROcode.addr = (char *) PROcode;
    ftn_char_PROcode.len = lenPROcode;

    ftn_char_ambname.addr = (char *) talkarray[indexCMB8ADfile];
    ftn_char_ambname.len = strlen(talkarray[indexCMB8ADfile]);

    ftn_char_AmbDirName.addr = (char *) AmbDirName;
    ftn_char_AmbDirName.len = strlen(AmbDirName);

    ftn_char_proname.addr = (char *) talkarray[indexCMB8profile];
    ftn_char_proname.len = strlen(talkarray[indexCMB8profile]);

    ftn_char_ProDirName.addr = (char *) ProDirName;
    ftn_char_ProDirName.len = strlen(ProDirName);

    ftn_char_SumDirName.addr = (char *) SumDirName;
    ftn_char_SumDirName.len = strlen(SumDirName);

    ftn_char_sizes.addr = (char *) sizearr;
    ftn_char_sizes.len = sizeof(sizearr[0]);;

    siz = nSPtot * nPROtot;
    error = getmem("atotal", (void **) &atotal, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    error = getmem("uatotal", (void **) &uatotal, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }


    siz = nSPtot;
    error = getmem("ambSP", (void **) &ambSP, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    error = getmem("uambSP", (void **) &uambSP, siz, __LINE__, 0);

    siz = nPRtot;

    error = getmem("inestim", (void **) &inestim, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    siz = nPRtot;
    error = getmem("sngval", (void **) &sngval, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    dimCwork = 10000;
    siz = dimCwork;
    error = getmem("Cwork", (void **) &Cwork, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    ftn_char_Cwork.addr = (char *) Cwork;
    ftn_char_Cwork.len = dimCwork;

    dimiwork1 = max(nSPprofile, nSPambient);
    dimiwork1 = max(nSPtot, dimiwork1);
    siz = dimiwork1;
    error = getmem("iwork1", (void **) &iwork1, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    dimwork = dimiwork1;
    siz = dimwork;
    error = getmem("CMBwork",  &CMBwork, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    error = getmem("work2", (void **) &work2, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    error =  INITIALIZE(
    &nSPtot,
    &nPRtot,
    &nADtot,
    &nPROtot, ftn_char_PROcode,
    &numsizes, ftn_char_sizes,
    ftn_char_ambname, ftn_char_AmbDirName,
    ftn_char_proname, ftn_char_ProDirName,
    ftn_char_SumDirName,
    atotal, uatotal,
    &iDummy);

    if (error < 0)
    {
      return error;
    }

//  Return totals to Delphi's (CMB82Internals.pas) Task 3 via direct calls:
    Lfp1 = (int *) w1;
    *Lfp1 =  nSPtot;

    Lfp2 = (int *) w2;
    *Lfp2 =  nPRtot;

    Lfp3 = (int *) w3;
    *Lfp3 =  nADtot;


// NB: methinks these 2 vars are's used by Delphi any more (TC; 11/07/04)
/*  Lfp4 = (int *) w4;
    *Lfp4 =  numsizes;

    Lfp5 = (int *) w5;
    *Lfp5 =  ambrecflag;
*/

    // allocate work arrays
    dimiwork1 = 2 * nSPtot + nPRtot;
    siz = dimiwork1;
    error = getmem("iwork1", (void **) &iwork1, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    dimCMBwork  = 9 * nSPtot;
    dimCMBwork += 7 * nPRtot;
    dimCMBwork += 2 * nSPtot * nPRtot;
    dimCMBwork += nPRtot * nPRtot;
    dimCMBwork += nSPtot * nSPtot;
    dimCMBwork += nSPtot * nSPtot; // additional storage for CMBSVD's workixi
    dimCMBwork += nSPtot;          // additional storage for CMBSVD's workix1
    dimCMBwork += nSPtot;          // additional storage for veffinv
    dimCMBwork += nSPtot * nSPtot; // storage for mPIN matrix
    dimCMBwork += nPRtot * nPRtot; // storage for inestimable sources

    siz =  dimCMBwork;
    error = getmem("CMBwork",  &CMBwork, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    siz = nSPtot;
    error = getmem("iSPfit", (void **) &iSPfit, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    siz = nPRtot;

    error = getmem("jPRfit", (void **) &jPRfit, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    dimerrbuf = 4 + nPRtot;
    lenerrbuf = LENERRBUF;
    siz = dimerrbuf * lenerrbuf + 1;

    error = getmem("errbuf", (void **) &errbuf, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    ftn_char_errbuf.addr = (char *) errbuf;
    ftn_char_errbuf.len = lenerrbuf;

    //We initialize outbuf here to avoid problems with calls to s.CMBSVD:
    dimoutbuf = 100;
    lenoutbuf = 80;
    siz = dimoutbuf * lenoutbuf + 1;

    error = getmem("outbuf", (void **) &outbuf, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    ftn_char_outbuf.addr = (char *) outbuf;
    ftn_char_outbuf.len = lenoutbuf;


    SFitLen = 180 + 24 * nPRtot + 1;
    siz = SFitLen;

    error = getmem("sfitbuf", (void **) &SFitBuf, siz, __LINE__, 0); // NB: sfitlen must be lower case to
	                                                                 // avoid runtie error; go figure -TC
    if (error < 0)
    {
      return error;
    }

    memset(SFitBuf, ' ', SFitLen);
    ftn_char_SFit.addr = (char *) SFitBuf;
    ftn_char_SFit.len = siz;

    return 0;
    break;


//  Step II; Delphi sends in 6 parameters
    case DO_SendCodes :
    cfp1 = (char *) w1;
    cfp2 = (char *) w2;
    cfp3 = (char *) w3;
//    cfp4 = (char *) w4;                    // See comments below ...
    cfp5 = (char *) w5;
    cfp6 = (char *) w6;

    strcpy(cfp1, (const char *) SPcode);
    strcpy(cfp2, (const char *) PRcode);
    strcpy(cfp3, (const char *) ADcode);
//    strcpy(cfp4, (const char *) sizearr); // sizeARR = 12-char field, e.g. 'FINE  COARSE'
                                            // On 11/15/04, TC suspected that SizeArray on 
                                            // the Delphi side isn't even used.  So I shut 
                                            // down the pass, and commented out the receive 
                                            // in CMB83Internals.pas' Task 3; works fine.
    strcpy(cfp5, (const char *) SPdefsel);
    strcpy(cfp6, (const char *) PRdefsel);

    break; // NB: no return from this step - an internal proc only ?


//  Step III (always invoked when Options button is pushed)
//  # decimals displayed variable only used in CMB82.c; NOT sent on to CMB82main.for.
    case DO_SetDisplayDecPl :

    Lfp1 = (int *) w1;
    COMMON_CMB.DisplayDecPl = *Lfp1;

    error = 0;
    return error;
    break;

//  Step IV  This module is invoked any time the sample selection array is 
//  changed to signal multiple samples (= batch mode processing; Batch Mode 
//  precludes user intervention), and when invoked, fRun sets batchFlag = action.  
//  It is called by CMB82Internal's SetBatchFlag, and is invoked only once before 
//  a calculation is attempted.
    case DO_SetBatchFlag :
    sel = DO_SetBatchFlag;

    Lfp1 = (int *) w1;
    action = *Lfp1;   // flag value: 0 => single sample; 1 => batch mode

    error = FRUN(
    &nSPtot, &nSPfit, ftn_char_SPcode, ftn_char_SPcode_user,
    &nPRtot, &nPRfit, ftn_char_PRcode, ftn_char_PRcode_user,
    &numCMBfit, &nPROtot, ftn_char_PROcode,
    &nADtot, ftn_char_ADcode,
    ambSP, uambSP,
    atotal, uatotal, sngval,
    iSPfit, jPRfit, inestim,
    &dimiwork1,iwork1,
    &dimCMBwork, CMBwork,
    &sel, &action, &action1,
    ftn_char_outbuf, &dimoutbuf, &lenoutbuf,
    ftn_char_Cwork,
    ftn_char_errbuf, &dimerrbuf, &lenerrbuf,
    &SFitbas, &uSFitbas, &sourcefitflag, &DichotFlagC);

    return error;
    break;


//  Step V (10/24/04; TC): This module explicitly passes a Dichot(=2)/Monochot(=1) 
	                    // flag to s.PData when fRun is called in Step X.
    case DO_SetDichotFlag :
    Lfp1        = (int *) w1;
    DichotFlagC = *Lfp1;

    error = 0;
    return error;
    break;


//  Step VI; Samples/Species/Sources selection tabs have been presented to user
//	        for review/adjustment; this step is invoked once per tagged sample.
    case DO_CMB :

    Lfp2 = (int *) w6;
    action = *Lfp2;

    Lfp3 = (int *) w7;
    action1 = *Lfp3;

    Lfp4 = (int *) w8;
    sourcefitflag = *Lfp4;

    // Initialize codes

    if ((action % 4 == 0) ||(action % 4 == 2))
    {
      cfp0 = (char *) w1;
      strncpy((char *) SPcode_user, cfp0, lenSPcode * nSPtot);

      cfp1 = (char *) w2;
      strncpy((char *) PRcode_user, cfp1, lenPRcode * nPRtot);

      cfp2 = (char *) w3; // sfp2 matches ADcode
      strncpy((char *) ADcode, cfp2, lenADcode * nADtot);

      error = SETAMBREC(ftn_char_ADcode); // Here, error is the (amb. sample) Record Number
      if (error < 0)
      {
        return error;
      }

    }

    sel = DO_CMB;

    error = FRUN(
    &nSPtot, &nSPfit, ftn_char_SPcode, ftn_char_SPcode_user,
    &nPRtot, &nPRfit, ftn_char_PRcode, ftn_char_PRcode_user,
    &numCMBfit, &nPROtot, ftn_char_PROcode,
    &nADtot, ftn_char_ADcode, // CHECK ftn_char_ADcode
    ambSP, uambSP,
    atotal, uatotal, sngval,
    iSPfit, jPRfit, inestim,
    &dimiwork1,iwork1,
    &dimCMBwork, CMBwork,
    &sel, &action, &action1,
    ftn_char_outbuf, &dimoutbuf, &lenoutbuf,
    ftn_char_Cwork,
    ftn_char_errbuf, &dimerrbuf, &lenerrbuf,
    &SFitbas, &uSFitbas, &sourcefitflag, &DichotFlagC);

    // Send error buffer to Delphi
    cfp4 = (char *) w5;
    strncpy(cfp4, (const char *) errbuf, dimerrbuf * LENERRBUF);

// Send measures to Delphi; DF (DegFree) is not included here cuz it's snagged from StrBuf.
//  cHat(1) has just been set in s.CMBSVD.  We hand off to sumSCEs.
//  It should now be passed to Task 10 of CMB82Internals.pas!

    ffp9  = (float *) w9;
    *ffp9 = COMMON_CMB.CHIsqr;

    ffp10  = (float *) w10;
    *ffp10 = COMMON_CMB.Rsquare;

    ffp11  = (float *) w11;
    *ffp11 = COMMON_CMB.pcntMass;

    ffp12  = (float *) w12;
    *ffp12 = COMMON_CMB.sumSCEs;

    return error; // Returned from fRun
    break;


//  Step VII
    case DO_SendFracEst :
    sel = DO_SendFracEst;

    ffp2 = (float *) w2;
    COMMON_CMB.maxSourceUnc = *ffp2;

    ffp3 = (float *) w3;
    COMMON_CMB.minSourceProj = *ffp3;

    error = FRUN(
    &nSPtot, &nSPfit, ftn_char_SPcode, ftn_char_SPcode_user,
    &nPRtot, &nPRfit, ftn_char_PRcode, ftn_char_PRcode_user,
    &numCMBfit, &nPROtot, ftn_char_PROcode,
    &nADtot, ftn_char_ADcode,
    ambSP, uambSP,
    atotal, uatotal, sngval,
    iSPfit, jPRfit, inestim,
    &dimiwork1,iwork1,
    &dimCMBwork, CMBwork,
    &sel, &action, &action1,
    ftn_char_outbuf, &dimoutbuf, &lenoutbuf,
    ftn_char_Cwork,
    ftn_char_errbuf, &dimerrbuf, &lenerrbuf,
    &SFitbas, &uSFitbas, &sourcefitflag, &DichotFlagC);

    ffp1 = (float *) w1;
    *ffp1 =  COMMON_CMB.fracestim;

    return error;
    break;


//  Step VIII
    case DO_SendInfo:
    sel = DO_SendInfo;

    error = FRUN(
    &nSPtot, &nSPfit, ftn_char_SPcode, ftn_char_SPcode_user,
    &nPRtot, &nPRfit, ftn_char_PRcode, ftn_char_PRcode_user,
    &numCMBfit, &nPROtot, ftn_char_PROcode,
    &nADtot, ftn_char_ADcode,
    ambSP, uambSP,
    atotal, uatotal, sngval,
    iSPfit, jPRfit, inestim,
    &dimiwork1,iwork1,
    &dimCMBwork, CMBwork,
    &sel, &action, &action1,
    ftn_char_outbuf, &dimoutbuf, &lenoutbuf,
    ftn_char_Cwork,
    ftn_char_errbuf, &dimerrbuf, &lenerrbuf,
    &SFitbas, &uSFitbas, &sourcefitflag, &DichotFlagC);

    Lfp1 = (int *) w1;
    *Lfp1 = nSPfit;

    Lfp2 = (int *) w2;
    *Lfp2 = nPRfit;

    Lfp3 = (int *) w3;
    *Lfp3 = numCMBfit;

    Lfp4 = (int *) w4;
    *Lfp4 = (int)(double) *(&ambSP[0]);

    return error;
    break;


//  Step IX
    case DO_SendSNGVAL:

    ffp1 = (float *) w1;
    /* return singular values to Delphi */
    for(n=0; n<nPRfit; n++)
    {
      *(ffp1+n) =  sngval[n];
    }

    return 0;
    break;


//  Step X    
    case DO_PData :

    Lfp1 = (int *) w1;
    lenoutbuf = *Lfp1;

    Lfp2 = (int *) w2;
    dimoutbuf = *Lfp2;

    Lfp3 = (int *) w3;
    sizoutbuf = *Lfp3;

    ffp5 = (float *) w5;
    COMMON_CMB.maxSourceUnc = *ffp5;

    ffp6 = (float *) w6;
    COMMON_CMB.minSourceProj = *ffp6;

    error = getmem("outbuf", (void **) &outbuf, (size_t) sizoutbuf, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    ftn_char_outbuf.addr = (char *) outbuf;
    ftn_char_outbuf.len = lenoutbuf;

    sel = DO_PData;

    error = FRUN(                           // 39 arguments
    &nSPtot, &nSPfit, ftn_char_SPcode, ftn_char_SPcode_user,
    &nPRtot, &nPRfit, ftn_char_PRcode, ftn_char_PRcode_user,
    &numCMBfit, &nPROtot, ftn_char_PROcode,
    &nADtot, ftn_char_ADcode,
    ambSP, uambSP,
    atotal, uatotal, sngval,
    iSPfit, jPRfit, inestim,
    &dimiwork1,iwork1,
    &dimCMBwork, CMBwork,
    &sel, &action, &action1,
    ftn_char_outbuf, &dimoutbuf, &lenoutbuf,
    ftn_char_Cwork,
    ftn_char_errbuf, &dimerrbuf, &lenerrbuf,
    &SFitbas, &uSFitbas, &sourcefitflag, &DichotFlagC);

    // Send output buffer to Delphi
    cfp3 = (char *) w4;
    strncpy(cfp3, (const char *) outbuf, sizoutbuf);

// TC added 10/30/04.  Send these 2 values to Delphi for Dichot. samples:
    ffp7  = (float *) w7;
    *ffp7 = COMMON_CMB.pcntMass;

    ffp8  = (float *) w8;
    *ffp8 = COMMON_CMB.sumSCEs; // THIS sumSCEs is set in s.PData {= work(ncHat+1)}
	                            // this value is for the TOTAL sumSCEs (Fine + Coarse)
    return error;
    break;


//  Step XI
    case DO_SScont :

    Lfp1 = (int *) w1;
    lenoutbuf = *Lfp1;

    Lfp2 = (int *) w2;
    dimoutbuf = *Lfp2;

    Lfp3 = (int *) w3;
    sizoutbuf = *Lfp3;

    error = getmem("outbuf", (void **) &outbuf, (size_t) sizoutbuf, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    ftn_char_outbuf.addr = (char *) outbuf;
    ftn_char_outbuf.len = lenoutbuf;

    dimCwork = lenoutbuf;

/*    if (error < 0)
    {
      return error;  ====>>>>> WHY IS THIS NEEDED?
    }
*/
    error = getmem("Cwork", (void **) &Cwork, dimCwork, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    ftn_char_Cwork.addr = (char *) Cwork;
    ftn_char_Cwork.len = dimCwork;

    sel = DO_SScont;

    error = FRUN(
    &nSPtot, &nSPfit, ftn_char_SPcode, ftn_char_SPcode_user,
    &nPRtot, &nPRfit, ftn_char_PRcode, ftn_char_PRcode_user,
    &numCMBfit, &nPROtot, ftn_char_PROcode,
    &nADtot, ftn_char_ADcode,
    ambSP, uambSP,
    atotal, uatotal, sngval,
    iSPfit, jPRfit, inestim,
    &dimiwork1,iwork1,
    &dimCMBwork, CMBwork,
    &sel, &action, &action1,
    ftn_char_outbuf, &dimoutbuf, &lenoutbuf,
    ftn_char_Cwork,
    ftn_char_errbuf, &dimerrbuf, &lenerrbuf,
    &SFitbas, &uSFitbas, &sourcefitflag, &DichotFlagC);

    // send output buffer to Delphi
    cfp3 = (char *) w4;
    strncpy(cfp3, (const char *) outbuf, sizoutbuf);

    return error;
    break;


//  Step XII
    case DO_mPIN :

    Lfp1 = (int *) w1;
    lenoutbuf = *Lfp1;

    Lfp2 = (int *) w2;
    dimoutbuf = *Lfp2;

    Lfp3 = (int *) w3;
    sizoutbuf = *Lfp3;

    error = getmem("outbuf", (void **) &outbuf, (size_t) sizoutbuf, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }

    ftn_char_outbuf.addr = (char *) outbuf;
    ftn_char_outbuf.len = lenoutbuf;

    dimCwork = lenoutbuf;
    siz = dimCwork;

    error = getmem("Cwork", (void **) &Cwork, siz, __LINE__, 0);
    if (error < 0)
    {
      return error;
    }


    ftn_char_Cwork.addr = (char *) Cwork;
    ftn_char_Cwork.len = dimCwork;

    sel = DO_mPIN;

    error = FRUN(
    &nSPtot, &nSPfit, ftn_char_SPcode, ftn_char_SPcode_user,
    &nPRtot, &nPRfit, ftn_char_PRcode, ftn_char_PRcode_user,
    &numCMBfit, &nPROtot, ftn_char_PROcode,
    &nADtot, ftn_char_ADcode,
    ambSP, uambSP,
    atotal, uatotal, sngval,
    iSPfit, jPRfit, inestim,
    &dimiwork1,iwork1,
    &dimCMBwork, CMBwork,
    &sel, &action, &action1,
    ftn_char_outbuf, &dimoutbuf, &lenoutbuf,
    ftn_char_Cwork,
    ftn_char_errbuf, &dimerrbuf, &lenerrbuf,
    &SFitbas, &uSFitbas, &sourcefitflag, &DichotFlagC);

    // send output buffer to Delphi
    cfp3 = (char *) w4;
    strncpy(cfp3, (const char *) outbuf, sizoutbuf);

    return error;
    break;


//  Step XIII
    case DO_SFitOut :

    Lfp1 = (int *) w1;
    Ltmp = *Lfp1;
    cfp2 = (char *) w2;
    strcpy((char *) SFitName, cfp2); // full path to $CMBoutDB.ext; SFitBuf is directed here
    Lfp3 = (int *) w3;
    Write2DT2status = *Lfp3;

    SFitRecLen = 123 + 20 * nPRtot; // Expanded front end of the output buffer file (from 104) 
	                                // to accomodate the 2 fields I added (i.e., Calc. concentration 
	                                // & Calculated uncertainty); 02/24/05 -TC

    if (Ltmp == 1 || Ltmp == 2) // output format is TXT or CSV
    {
      error = SFitOut_c(&Ltmp);
    }
    else if (Ltmp == 3)         // output format is DBF
    {
      error = SFitDBF_c();
    }
    else if (Ltmp == 4)         // output format is WKS
    {
      error = SFitWKS_c();
    }
    else
    {
      error = -12;
    }
    return error;
    break;


    case DO_CSV2TXT :

    cfp1 = (char *) w9;
    strncpy((char *) infile, cfp1, MAXFILELENGTH);

    cfp2 = (char *) w10;
    strncpy((char *) outfile, cfp2, MAXFILELENGTH);

    error = csv2txt(infile, outfile);

    return error;
    break;


    case DO_DBF2TXT :

    cfp1 = (char *) w9;
    strncpy((char *) infile, cfp1, MAXFILELENGTH);

    cfp2 = (char *) w10;
    strncpy((char *) outfile, cfp2, MAXFILELENGTH);

    error = dbf2txt(infile, outfile);

    return error;
    break;


    case DO_WKS2TXT :

    cfp1 = (char *) w9;
    strncpy((char *) infile, cfp1, MAXFILELENGTH);

    cfp2 = (char *) w10;
    strncpy((char *) outfile, cfp2, MAXFILELENGTH);

    error = wks2txt(infile, outfile);

    return error;
    break;


    case DO_CAR2TXT :

    cfp1 = (char *) w9;
    strncpy((char *) infile, cfp1, MAXFILELENGTH);

    cfp2 = (char *) w10;
    strncpy((char *) outfile, cfp2, MAXFILELENGTH);

    error = car2txt(infile, outfile);

    return error;
    break;


//  Step XIV: Termination module
    case DO_CMB_Terminate :

    // free all memory
    error = getmem("talkarray", (void **) &talkarray, siz, __LINE__, 3);

    remove(ProDirName); // EPA-CMB8.2 cleans up after itself!
    remove(AmbDirName);
	remove(SumDirName);

    return error;
    break;

    }
    fcloseall();
    return 0;
}

// =============================================================================
int setSPcode()
{
  char *seps = " \t";
  size_t sizCwork;
  size_t sizeSPcode, siz;
  int error, n, nSP, ntmp, len, nn;
  int flag;
  char *cp;
  int ntok, ntok1;

  sizCwork = 10000;
  error = getmem("Cwork", (void **) &Cwork, sizCwork, __LINE__, 0);
  error = getmem("Cwork1", (void **) &Cwork1, sizCwork, __LINE__, 0);

  if (error < 0)
  {
    goto Error_Handler;
  }

  // Set up temporary ambient data array; 
  // NB: this assumes that the required AD* file has already been selected and opened!
  // If (for whatever reason) it hasn't been, CMB82.c will choke here!  Big time. -TC
  pFile = fopen(talkarray[indexCMB8ADfile], "r");
  if (!pFile)
  {
    sprintf(Cwork1, "Can't open CMB8.2 ambient sample data file: %s",
    talkarray[indexCMB8ADfile]);
    message(messagetitle, Cwork1);
    error = ERROR_FILE_OPEN;
    goto Error_Handler;
  }

  // Determine maximum # species - nSPtot
  // Norm R - need to pick up species from ambient data file, since organics don't have size TOTAL

  nSPtot = 0;
  while(1) // Find nonblank lines and count number of tokens
  {
    cp = fgets(Cwork1, sizCwork, pFile);
    if (cp == NULL) break;
    n = tokcnt(Cwork1);

    if (n > 5)
    {
      ntmp = n - 5;

      if ((ntmp % 2) != 0)
      {
        sprintf(winbuf,
        "Incorrect number of tokens in header line of file %s.  Should be an odd number > 5!",
        talkarray[indexCMB8ADfile]);
        message(messagetitle, winbuf);
        error = ERROR_AMBIENT_SET;
        goto Error_Handler;
      }

      // NB - we are assumming here that the first measurement is TOTAL
      nSPtot = ntmp / 2;

      // Set up temporary element array
      sizeSPcode = nSPtot+1;

      error = getmem("SPcode", (void **) &SPcode, sizeSPcode, __LINE__, 0);
      if (error < 0)
      {
         goto Error_Handler;
      }


      strcpy(cbuf, strtok(Cwork1, seps)); // first token - Profile code

      strcpy(cbuf, strtok(NULL, seps));
      strcpy(cbuf, strtok(NULL, seps));
      strcpy(cbuf, strtok(NULL, seps));
      strcpy(cbuf, strtok(NULL, seps));

      memset(SPcode, ' ', lenSPcode * nSPtot);

      for(n=0; n<nSPtot; n++)
      {
        strcpy(cbuf, strtok(NULL, seps));
        Ljust1(cbuf, lenSPcode1);
        sprintf(SPcode[n], "%s", cbuf);
        SPcode[n][lenSPcode1] = ' ';
        strcpy(cbuf, strtok(NULL, seps));
        Ljust1(cbuf, lenSPcode2);
        sprintf(SPcode[n] + lenSPcode1+1, "%s", cbuf);
        SPcode[n][lenSPcode1+lenSPcode2+1] = ' ';

      }

      break;
    }

  }


  if (nSPtot <= 0)
  {
    sprintf(Cwork1, "Total number of species = %ld", nSPtot);
    message(messagetitle, Cwork1);
    error = ERROR_SPECIES_SET;
    goto Error_Handler;
  }

  SPcode[nSPtot][0] = '\0'; // termination null

  // Now compare with source profiles data file

  fclose(pFile);

  pFile = fopen(talkarray[indexCMB8profile], "r");
  if (!pFile)
  {
    sprintf(Cwork1, "Can't open CMB8.2 source profiles data file: %s",
    talkarray[indexCMB8profile]);
    message(messagetitle, Cwork1);
    error = ERROR_FILE_OPEN;
    goto Error_Handler;
  }


  flag = 1;
  while((cp = fgets(Cwork1, sizCwork, pFile)))
  {
    n = tokcnt(Cwork1);

    if (n > 3)
    {
      ntok1 = StuffTokArray((char *) tokens1, MAXTOKENS, LENTOKEN,  Cwork1);
      break;
    }

  }

  for(nSP=1; nSP<nSPtot; nSP++) //skip TOTAL
  {
    strncpy(cbuf1, SPcode[nSP], lenSPcode1);
    cbuf1[lenSPcode1] = '\0';

    // Match with species codes in source profiles (PR) file

    rewind(pFile);
    flag = 1;
    for(ntok=3; ntok<ntok1; ntok = ntok + 2)
    {

      {

        if (!StrCmpNR(cbuf1, tokens1[ntok]))
        {
          flag = 0;  // Match
          break;
        }

      }

    }

    if (flag) // Remove species for SPcode
    {
      for(n=nSP; n<nSPtot-1; n++)
      {
        strncpy(SPcode[n], SPcode[n+1], lenSPcode);
      }

      nSPtot--;
      SPcode[nSPtot][0] = '\0';
      nSP--;
    }

  }

  if (nSPtot <= 0)
  {
    sprintf(Cwork1, "Total number of species = %ld", nSPtot);
    message(messagetitle, Cwork1);
    error = ERROR_SPECIES_SET;
    goto Error_Handler;
  }

  fclose(pFile);

  // Now examine the species (SP) selection file if it's being used:
  siz = nSPtot+1;

  error = getmem("SPdefsel", (void **) &SPdefsel, siz, __LINE__, 0);
  if (error < 0)
  {
    goto Error_Handler;
  }

  memset(SPdefsel, ' ', nSPtot*LenDefSel);
  strcpy(SPdefsel[0], "T T T T T T T T T T"); // TOTAL
  SPdefsel[nSPtot][0] = '\0';


  // NB: If fopen fails, it returns a NULL pointer and the file is never opened,
  //     so it's not necessary to call fclose() on pFile (at end of this function).
  pFile = fopen(talkarray[indexCMB8SPselect], "r");
  if (pFile)
  {

    for(nSP=1; nSP<nSPtot; nSP++) // Skip TOTAL
    {
      strncpy(cbuf1, SPcode[nSP], lenSPcode1);
      cbuf1[lenSPcode1] = '\0';	  // Tack null on the end

      // Match with species codes in species selection file

      rewind(pFile);              // Move file pointer back to the top of file (ToF)
      flag = 1;
      while((cp = fgets(Cwork1, sizCwork, pFile)))
      {
        n = tokcnt(Cwork1);

        if (n > 1)
        {
          strcpy(Cwork, Cwork1);
          strcpy(cbuf2, strtok(Cwork, seps)); // first token - Profile code

          if (!StrCmpNR(cbuf1, cbuf2))
          {
            flag = 0;                         // Found in first species list

            // Set selection *
            len = strlen(Cwork1);
            if (len >= 19)
            {
              if (Cwork1[18] == '*')
              {
                SPcode[nSP][16] = '*';
                SPcode[nSP][18] = '*';
                SPcode[nSP][20] = '*';
                SPcode[nSP][22] = '*';
              }

              // Initial selection
              SPdefsel[nSP][0] = Cwork1[18];

              for(nn=1; nn<10; nn++)
              {
                if (len >= 19 + 2*nn)
                {
                  // Initial selection
                  if (Cwork1[18+2*nn] == '*')
                  {
                    SPdefsel[nSP][2*nn] = '*';
                  }
                }
              }
            }

            break;
          }

        }

      }

      if (flag) // Remove species for SPcode
      {
        for(n=nSP; n<nSPtot-1; n++)
        {
          strncpy(SPcode[n], SPcode[n+1], lenSPcode);
        }

        nSPtot--;
        SPcode[nSPtot][0] = '\0';
        SPdefsel[nSPtot][0] = '\0';
        nSP--;
      }

    }

    if (nSPtot <= 0)
    {
      sprintf(Cwork1, "Total number of species = %ld", nSPtot);
      message(messagetitle, Cwork1);
      error = ERROR_SPECIES_SET;
      goto Error_Handler;
    }

    fclose(pFile);
  }

  error = 0;
  return error; // TC/JS added 08-19-04 - exit out before we hit error handler 
                // code, which we only want to do if the file was actually open.
                // Since the file may not be open, we should just exit rather 
                // than calling fclose() on pFile, which could be a null 
                // pointer and will result in an exception being thrown.

Error_Handler:
  fclose(pFile);
  return error;

}

// =============================================================================
int setPRcode()
{

  FILE *pFile;

  size_t sizCwork;
  size_t sizPRcode;
  size_t siztoks;

  char *cp;
  int n, ntmp, ntot, flag;
  int nPR, nn;
  int count;
  int ntmp1, ntmp2;

  int len;

  char *seps = " \t";
  int ntok1;
  int error;

  static char(*tokens1) [LENTOKEN] = NULL;
  static char(*tokens2) [LENTOKEN] = NULL;
  int maxtoks;
  int numtoks, NUMTOKS, nline;

  #define MODE777 0777

  sizCwork = 10000;

  error = getmem("Cwork", (void **) &Cwork, sizCwork, __LINE__, 0);
  if (error < 0)
  {
    error = ERROR_MEMORY_ALLOC;
    goto Error_Handler;
  }

  error = getmem("Cwork1", (void **) &Cwork1, sizCwork, __LINE__, 0);
  if (error < 0)
  {
    error = ERROR_MEMORY_ALLOC;
    goto Error_Handler;
  }

  error = getmem("Cwork1", (void **) &Cwork1, sizCwork, __LINE__, 0);
  if (error < 0)
  {
    error = ERROR_MEMORY_ALLOC;
    goto Error_Handler;
  }


  pFile = fopen(talkarray[indexCMB8profile], "r");
  if (!pFile)
  {
    sprintf(Cwork1, "Can't open CMB8.2 source profiles data file: %s",
    talkarray[indexCMB8profile]);
    message(messagetitle, Cwork1);
    error = ERROR_FILE_OPEN;
    goto Error_Handler;
  }

  // Count number of nonblank lines and set maximum total number of proient data records
  nPRtot = 0;
  flag = 1;
  count = 0;
  while((cp = fgets(Cwork1, sizCwork, pFile)))
  {

    if ((n = tokcnt(Cwork1)) > 3)
    {
      if (count == 0)
      {
        count = 1;
        NUMTOKS = n;
        maxtoks = 2 * n; // Allow for data file errors resulting in too many tokens
        siztoks = maxtoks;

        error = getmem("tokens1", (void **) &tokens1, siztoks, __LINE__, 0);
        if (error)
        {
          sprintf(winbuf, "Memory allocation error");
          message(messagetitle, winbuf);
          error = ERROR_MEMORY_ALLOC;
          goto Error_Handler;
        }

        error = getmem("tokens2", (void **) &tokens1, siztoks, __LINE__, 0);
        if (error)
        {
          sprintf(winbuf, "Memory allocation error");
          message(messagetitle, winbuf);
          error = ERROR_MEMORY_ALLOC;
          goto Error_Handler;
        }
      }
      else
      {
        nPRtot++;
      }
    }

  }

  sizPRcode = nPRtot + 1;
  error = getmem("PRcode", (void **) &PRcode, sizPRcode, __LINE__, 0);
  if (error)
  {
    sprintf(winbuf, "Memory allocation error");
    message(messagetitle, winbuf);
    error = ERROR_MEMORY_ALLOC;
    goto Error_Handler;
  }

  rewind(pFile);  // Set up temporary proient data buffer
  ntot = nPRtot;
  nPRtot = 0;
  count = 0;
  nline = 0;
  while(nPRtot < ntot)
  {
    cp = fgets(Cwork1, sizCwork, pFile);
    if (cp == NULL) break;
    numtoks = StuffTokArray((char *) tokens1, maxtoks, LENTOKEN,  Cwork1);  // General data record
    nline++;
    if (numtoks > 3)
    {
      if (count == 0)
      {
        count = 1; // Skip header line
      }
      else
      {

        if (numtoks != NUMTOKS)
        {
          sprintf(winbuf,
          "Number of tokens in line %d of file %s not equal to number of tokens in header line!",
          nline, talkarray[indexCMB8profile]);
          message(messagetitle, winbuf);
          error = ERROR_SOURCES_SET;
          goto Error_Handler;
        }

        flag = 1;
        for(nPR=0; nPR<nPRtot; nPR++)
        {
          strncpy(cbuf, PRcode[nPR], lenPRcode1);

          if (!StrCmpNR(cbuf, tokens1[0]))
          { // Match
            flag = 0;
            break;
          }
        }

        if (flag) // New source
        {
          memset(PRcode[nPRtot], ' ', lenPRcode);

          ntmp1 = 0;
          ntok1 = 0;
          ntmp = lenPRcode1;
          ntmp2 = ntmp1 + ntmp;

          strcpy(cbuf, tokens1[ntok1]);
          Ljust1(cbuf, ntmp+1); // Adds blank at end
          strncpy(cbuf1 + ntmp1, cbuf, ntmp+1);

          ntmp = lenPRcode2;
          ntmp1 = ntmp2+1;
          ntmp2 = ntmp1 + ntmp;
          ntok1++;
          strcpy(cbuf, tokens1[ntok1]);
          Ljust1(cbuf, ntmp+1);
          strncpy(cbuf1 + ntmp1, cbuf, ntmp+1);
          strncpy(PRcode[nPRtot], cbuf1, ntmp2+1);

          nPRtot++;

        }
        count++;

      }
    }

  }

  if (nPRtot <=0)
  {
    sprintf(Cwork1, "Total number of sources = %ld", nPRtot);
    message(messagetitle, Cwork1);
    error = ERROR_SOURCES_SET;
    goto Error_Handler;
  }

  // termination NULL
  PRcode[nPRtot][0] = '\0';

  fclose(pFile);

  // Now examine the source profiles (PR) selection file if it's being used:
  sizPRcode = nPRtot+1;

  error = getmem("PRdefsel", (void **) &PRdefsel, sizPRcode, __LINE__, 0);
  if (error < 0)
  {
    error = ERROR_MEMORY_ALLOC;
    goto Error_Handler;
  }

  memset(PRdefsel, ' ', nPRtot*LenDefSel);
  PRdefsel[nPRtot][0] = '\0';

  // NB: If fopen fails, it returns a NULL pointer and the file is never opened,
  //     so it's not necessary to call fclose() on pFile (at end of this function).
  pFile = fopen(talkarray[indexCMB8PRselect], "r");
  if (pFile != NULL)
  {

    for(nPR = 0; nPR < nPRtot; nPR++)
    {
      strncpy(cbuf1, PRcode[nPR], lenPRcode1);
      cbuf1[lenPRcode1] = '\0';	// Tack null onto the end

      // Match with species codes in species selection file

      rewind(pFile);	        // Move file pointer back to the top of file (ToF)
      flag = 1;

	  // read a lines from the file
      while((cp = fgets(Cwork1, sizCwork, pFile)))
      {
        n = tokcnt(Cwork1);

        if (n > 1)
        {
          strcpy(Cwork, Cwork1);
          strcpy(cbuf2, strtok(Cwork, seps)); // first token - Profile code

          if (!StrCmpNR(cbuf1, cbuf2))
          {
            flag = 0;  // Found in first species list

            // Set selection *
            len = strlen(Cwork1);
            if (len >= 19)
            {
              if (Cwork1[18] == '*')
              {
                PRcode[nPR][16] = '*';
                PRcode[nPR][18] = '*';
                PRcode[nPR][20] = '*';
                PRcode[nPR][22] = '*';
              }

              // Initial selection
              PRdefsel[nPR][0] = Cwork1[18];

              for(nn=1; nn<10; nn++)
              {
                if (len >= 19 + 2*nn)
                {
                  // Initial selection
                  if (Cwork1[18+2*nn] == '*')
                  {
                    PRdefsel[nPR][2*nn] = '*';
                  }
                }
              }
            }

            break;
          }

        }

      }

      if (flag) // Remove species for PRcode
      {
        for(n=nPR; n<nPRtot-1; n++)
        {
          strncpy(PRcode[n], PRcode[n+1], lenPRcode);
        }

        nPRtot--;
        PRcode[nPRtot][0] = '\0';
        PRdefsel[nPRtot][0] = '\0';
        nPR--;
      }

    }

    if (nPRtot <= 0)
    {
      sprintf(Cwork1, "Total number of sources = %ld", nPRtot);
      message(messagetitle, Cwork1);
//      error = ERROR_SPECIES_SET;  // TC corrected (line below) on 05-09-03
      error = ERROR_SOURCES_SET;
      goto Error_Handler; // Error_Handler
    }

    fclose(pFile);
  }

  error = 0;
  return error; // TC/JS added 08-19-04 - exit out before we hit error handler
                // code, which we only want to do if the file was actually open.
                // Since the file may not be open, we should just exit rather 
                // than calling fclose() on pFile, which could be a null 
                // pointer and will result in an exception being thrown.

  Error_Handler:
  fclose(pFile);
  return error;

}

// =============================================================================
int setADcode() // Creates and populates AMDdirect.dat; creates ADcode (char. string)
{

  FILE *pFile;
  int handle;

  size_t sizCwork;
  size_t sizambient;
  size_t siztoks;

  char *cp;
  int j, n, ntmp, ntot, flag;
  int count;
  int ntmp1, ntmp2;

  int nSP;

  float ftmp;
  char *seps = " \t";
  int ntok1, nAD;
  int error;

  static char(*tokens1) [LENTOKEN] = NULL;
  static char(*tokens2) [LENTOKEN] = NULL;
  int maxtoks;
  int numtoks, NUMTOKS, nline;

  #define MODE777 0777

  sizCwork = 10000;

  error = getmem("Cwork1", (void **) &Cwork1, sizCwork, __LINE__, 0);
  if (error < 0)
  {
    return error;
  }


  pFile = fopen(talkarray[indexCMB8ADfile], "r");
  if (!pFile)
  {
    sprintf(Cwork1, "Can't open CMB8.2 ambient sample data file: %s",
    talkarray[indexCMB8ADfile]);
    message(messagetitle, Cwork1);
    error = ERROR_FILE_OPEN;
    goto EXIT_ERROR;
  }

  // Ambient data direct access file:
  handle = open(AmbDirName, O_CREAT | O_BINARY | O_RDWR, MODE777);
  if (handle == -1)
  {
    sprintf(winbuf, "Can't create CMB8.2 direct access temporary file");
    message(messagetitle, winbuf);
    error = ERROR_FILE_OPEN;
    goto EXIT_ERROR;
  }

  error = lseek(handle, 0, SEEK_SET);

  // Count number of nonblank lines and set maximum total number of ambient data (AD) records
  nADtot = 0;
  flag = 1;
  count = 0;
  while((cp = fgets(Cwork1, sizCwork, pFile)))
  {

    if ((n = tokcnt(Cwork1)) > 5)
    {
      if (count == 0)
      {
        count = 1;
        NUMTOKS = n;
        maxtoks = 2 * n; // Allow for data file errors resulting in too many tokens
        siztoks = maxtoks;

        error = getmem("tokens1", (void **) &tokens1, siztoks, __LINE__, 0);
        if (error)
        {
          sprintf(winbuf, "Memory allocation error");
          message(messagetitle, winbuf);
          error = ERROR_MEMORY_ALLOC;
          goto EXIT_ERROR;
        }

        error = getmem("tokens2", (void **) &tokens2, siztoks, __LINE__, 0);
        if (error)
        {
          sprintf(winbuf, "Memory allocation error");
          message(messagetitle, winbuf);
          error = ERROR_MEMORY_ALLOC;
          goto EXIT_ERROR;
        }
      }
      else
      {
        nADtot++;
      }
    }
  }

  sizambient = nADtot + 1;
  error = getmem("ADcode", (void **) &ADcode, sizambient, __LINE__, 0);
  if (error)
  {
    sprintf(winbuf, "Memory allocation error");
    message(messagetitle, winbuf);
    error = ERROR_MEMORY_ALLOC;
    goto EXIT_ERROR;
  }

  rewind(pFile); // Set up temporary ambient data buffer for internal read(s):  ADcode
  ntot = nADtot;
  nADtot = 0;
  count = 0;
  nline = 0;
  while(nADtot < ntot)
  {
    cp = fgets(Cwork1, sizCwork, pFile); // cp loads Cwork1 (Header for AD*.* ambient data input file)
    if (cp == NULL) break;
    numtoks = StuffTokArray((char *) tokens1, maxtoks, LENTOKEN,  Cwork1);  // General data record
    nline++;
    if (numtoks > 5)
    {
      if (count == 0)
      {
        count = 1;                                                        // Skip header line
        n = StuffTokArray((char *) tokens2, maxtoks, LENTOKEN,  Cwork1);  // Record header tokens
      }
      else
      {

        if (numtoks != NUMTOKS)
        {
          sprintf(winbuf,
          "Number of tokens in line %d of file %s not equal to number of tokens in header line!",
          nline, talkarray[indexCMB8ADfile]);
          message(messagetitle, winbuf);
          error = ERROR_AMBIENT_SET;
          goto EXIT_ERROR;
        }

        memset(ADcode[nADtot], ' ', lenADcode);

        ntmp1 = 0;
        ntok1 = 0;
        ntmp = lenADcode1;
        ntmp2 = ntmp1 + ntmp;

        strcpy(cbuf, tokens1[ntok1]);
        Ljust1(cbuf, ntmp+1);                    // Adds blank at end
        strncpy(cbuf1 + ntmp1, cbuf, ntmp+1);
        error = write(handle, cbuf, lenADcode1); // Sample ID

        ntmp = lenADcode2;
        ntmp1 = ntmp2+1;
        ntmp2 = ntmp1 + ntmp;
        ntok1++;
        strcpy(cbuf, tokens1[ntok1]);
        Ljust1(cbuf, ntmp+1);
        strncpy(cbuf1 + ntmp1, cbuf, ntmp+1);
        error = write(handle, cbuf, lenADcode2); // Date

        ntmp = lenADcode3;
        ntmp1 = ntmp2+1;
        ntmp2 = ntmp1 + ntmp;
        ntok1++;
        strcpy(cbuf, tokens1[ntok1]);
        Ljust1(cbuf, ntmp+1);
        strncpy(cbuf1 + ntmp1, cbuf, ntmp+1);
        error = write(handle, cbuf, lenADcode3); // Duration

        ntmp = lenADcode4;
        ntmp1 = ntmp2+1;
        ntmp2 = ntmp1 + ntmp;
        ntok1++;
        strcpy(cbuf, tokens1[ntok1]);
        Ljust1(cbuf, ntmp+1);
        strncpy(cbuf1 + ntmp1, cbuf, ntmp+1);
        error = write(handle, cbuf, lenADcode4); // Start hour

        ntmp = lenADcode5;
        ntmp1 = ntmp2+1;
        ntmp2 = ntmp1 + ntmp;
        ntok1++;
        strcpy(cbuf, tokens1[ntok1]);
        Ljust1(cbuf, ntmp + 2);
        strncpy(cbuf1 + ntmp1, cbuf, ntmp+2);
        error = write(handle, cbuf, lenADcode5); // Size

        //record number
        ntmp = lenADcode6;
        ntmp1 = ntmp2+1;
        ntmp2 = ntmp1 + ntmp;
        sprintf(cbuf, "%06d", count);            // Record number in AMBdirect.dat
        strncpy(cbuf1 + 36, cbuf, lenADcode6);   // Critical change from 35 (6-char size fraction field)
        strncpy(ADcode[nADtot], cbuf1, lenADcode);

        for(nSP=0; nSP<nSPtot; nSP++)            // TOTAL INcluded
        {
          strncpy(cbuf, SPcode[nSP], lenSPcode1);
          cbuf[lenSPcode1] = '\0';

          for(n = 5; n<numtoks; n=n+2)
          {
            if (!StrCmpNR(cbuf, tokens2[n]))
            { //match
              strcpy(cbuf1, tokens1[n]);         // Measurement
              ftmp = (float) atof(cbuf1);
              error = write(handle, &ftmp, 4);

              strcpy(cbuf1, tokens1[n+1]);       // Uncertainty
              ftmp = (float) atof(cbuf1);
              error = write(handle, &ftmp, 4);
              break;
            }
          }
        }

        nADtot++;
        count++;

      }
    }
  }

  if (nADtot <=0)
  {
    sprintf(Cwork1, "Total number of ambient data records = %ld", nADtot); // TC corrected 08-19-04; changed nPRtot to nADtot
    message(messagetitle, Cwork1);
    error = ERROR_AMBIENT_SET;
    goto EXIT_ERROR;
  }

  // termination NULL
  ADcode[nADtot][0] = '\0';

  // Now load sizes array
  numsizes = 0;
  for(n=0; n<nADtot; n++)
  {
    strncpy(cbuf, ADcode[n]+INDEXSIZENAME, LENSIZENAME);

    flag = 1;
    for(j=0; j<numsizes; j++)
    {
      if (!strncmp(cbuf, sizearr[j], LENSIZENAME))
      {
        flag = 0;
        break;
      }
    }

    if (flag && numsizes < MAXSIZES)
    {
      strncpy(sizearr[numsizes], cbuf, LENSIZENAME);
      numsizes++;
    }

  }

  fclose(pFile);

// Now examine the ambient data (AD) selection file if it's being used.
// NB: If fopen fails, it returns a NULL pointer and the file is never opened,
//     so it's not necessary to call fclose() on pFile (at end of this function).

  pFile = fopen(talkarray[indexCMB8ADselect], "r");

  if (pFile)
  {

    for(nAD=0; nAD<nADtot; nAD++)
    {

      ParseAmbient((char *) tokens1, LENTOKEN, (char *) ADcode[nAD]);

      rewind(pFile); // Move file pointer back to the top of file (ToF)
      flag = 1;

      while((cp = fgets(Cwork1, sizCwork, pFile)))
      {
        n = tokcnt(Cwork1);

        if (n > 0)
        {
          ntok1 = StuffTokArray((char *) tokens2, MAXTOKENS, LENTOKEN,  Cwork1);


          if (ntok1 >= 5)
          {
            // Match codes
            if
     (
            !StrCmpNR(tokens1[0], tokens2[0]) &&    // Site ID
            !StrCmpNR(tokens1[1], tokens2[1]) &&    // Date
            atoi(tokens1[2]) == atoi(tokens2[2]) && // Duration
            atoi(tokens1[3]) == atoi(tokens2[3]) && // Start hour
            !StrCmpNR(tokens1[4], tokens2[4])       // Size
  )
            {
              flag = 0;

              // Set selection here

              if (ntok1 >= 6)
              {
                if (tokens2[5][0] == '*')
                {
                  ADcode[nAD][35] = '*'; // TC changed from 34 to 35 (11/07/04).  This explicitly places * in col. 36 in ADcode
                }
              }

              break;

            }
          }
        }
      }

      if (flag)
      {
        // Remove ambient data record
        for(n=nAD; n<nADtot-1; n++)
        {
          strncpy(ADcode[n], ADcode[n+1], lenADcode);
        }
        nADtot--;
        ADcode[nADtot][0] = '\0';
                nAD--;
      }

    }

    fclose(pFile);
  }

  if (nADtot <=0)
  {
    sprintf(Cwork1, "Total number of ambient data records = %ld", nADtot); // TC corrected 08-19-04; changed nPRtot to nADtot
    message(messagetitle, Cwork1);
    return ERROR_AMBIENT_SET;
  }

  error = 0;
  EXIT_ERROR:

  fcloseall();
  close(handle);

  return error;

}

// =============================================================================
int setPROcode() // Creates and populates PROdirect.dat; creates PROcode (char. string)

{
    char const *fileName;
    FILE *pFile;

  int handle;

  size_t sizCwork;
  size_t sizePROcode;
  size_t siztoks;

  char *cp;
  int n, ntot, flag;
  int count;
  int nSP;

  float ftmp;
  int ntok1;
  int error;


  static char(*tokens1) [LENTOKEN] = NULL;
  static char(*tokens2) [LENTOKEN] = NULL;
  int maxtoks;
  int numtoks, NUMTOKS, nline;
  int nbegin, nwrite;

  #define MODE777 0777

  sizCwork = 10000;

  error = getmem("Cwork1", (void **) &Cwork1, sizCwork, __LINE__, 0);
  if (error < 0)
  {
    return error;
  }

    fileName = talkarray[indexCMB8profile];
    pFile = fopen(fileName, "r");
    if (!pFile)
  {
    sprintf(Cwork1, "Can't open CMB8.2 source profiles data file: %s",
    talkarray[indexCMB8profile]);
    message(messagetitle, Cwork1);
    error = ERROR_FILE_OPEN;
    goto EXIT_ERROR;
  }

  // Proient data direct access file
  handle = open(ProDirName, O_CREAT | O_BINARY | O_RDWR, MODE777);
  if (handle == -1)
  {
    sprintf(winbuf, "Can't create CMB8.2 direct access temporary file");
    message(messagetitle, winbuf);
    error = ERROR_FILE_OPEN;
    goto EXIT_ERROR;
  }

  error = lseek(handle, 0, SEEK_SET);

  // Count number of nonblank lines and set maximum total number of profiles data records
  nPROtot = 0;
  flag = 1;
  count = 0;
  while((cp = fgets(Cwork1, sizCwork, pFile)))
  {

    if ((n = tokcnt(Cwork1)) > 5)
    {
      if (count == 0)
      {
        count = 1;
        NUMTOKS = n;
        maxtoks = 2 * n; // Allow for data file errors resulting in too many tokens
        siztoks = maxtoks;

        error = getmem("tokens1", (void **) &tokens1, siztoks, __LINE__, 0);
        if (error)
        {
          sprintf(winbuf, "Memory allocation error");
          message(messagetitle, winbuf);
          error = ERROR_MEMORY_ALLOC;
          goto EXIT_ERROR;
        }

        error = getmem("tokens2", (void **) &tokens2, siztoks, __LINE__, 0);
        if (error)
        {
          sprintf(winbuf, "Memory allocation error");
          message(messagetitle, winbuf);
          error = ERROR_MEMORY_ALLOC;
          goto EXIT_ERROR;
        }
      }
      else
      {
        nPROtot++;
      }
    }

  }

  sizePROcode = nPROtot + 1;

  error = getmem("PROcode", (void **) &PROcode, sizePROcode, __LINE__, 0);
  if (error)
  {
    sprintf(winbuf, "Memory allocation error");
    message(messagetitle, winbuf);
    error = ERROR_MEMORY_ALLOC;
    goto EXIT_ERROR;
  }

  rewind(pFile);  // Set up temporary PRofiles data buffer for internal read(s): PROcode
  ntot = nPROtot;
  nPROtot = 0;
  count = 0;
  nline = 0;
  while(nPROtot < ntot)
  {
    cp = fgets(Cwork1, sizCwork, pFile);
    if (cp == NULL) break;
    numtoks = StuffTokArray((char *) tokens1, maxtoks, LENTOKEN,  Cwork1);     // General data record
    nline++;
    if (numtoks > 3)
    {
      if (count == 0)
      {
        count = 1;                                                              // Skip header line
        numtoks = StuffTokArray((char *) tokens2, maxtoks, LENTOKEN,  Cwork1);  // General data record
      }
      else
      {

        memset(PROcode[nPROtot], ' ', lenPROcode);

        nbegin = 0;
        ntok1 = 0;
        nwrite = lenPROcode1;

        strcpy(cbuf, tokens1[ntok1]);
        Ljust1(cbuf, nwrite);                // Adds blank at end
        strncpy(cbuf1 + nbegin, cbuf, nwrite);
        error = write(handle, cbuf, nwrite); // Source code

        nbegin = nbegin + nwrite;
                nwrite = lenPROcode2;
        ntok1++;
        strcpy(cbuf, tokens1[ntok1]);
        Ljust1(cbuf, nwrite);                // Adds blank at end
        strncpy(cbuf1 + nbegin, cbuf, nwrite);
        error = write(handle, cbuf, nwrite); // Source name

        nbegin = nbegin + nwrite;
                nwrite = lenPROcode3;
        ntok1++;
        strcpy(cbuf, tokens1[ntok1]);
        Ljust1(cbuf, nwrite);                // Adds blank at end
        strncpy(cbuf1 + nbegin, cbuf, nwrite);
        error = write(handle, cbuf, nwrite); // Size


        // Record number

        nbegin = nbegin + nwrite;
                nwrite = lenPROcode4;
        ntok1++;
        sprintf(cbuf, "%06d", count);        // record number in PROdirect.dat
        strncpy(cbuf1 + nbegin, cbuf, nwrite);

        strncpy(PROcode[nPROtot], cbuf1, lenPROcode);

        ftmp = 1.; // Total
        error = write(handle, &ftmp, 4);
        ftmp = 0.;
        error = write(handle, &ftmp, 4);

        for(nSP=1; nSP<nSPtot; nSP++)        // TOTAL EXcluded
        {
          strncpy(cbuf, SPcode[nSP], lenSPcode1);
          cbuf[lenSPcode1] = '\0';

          for(n = 3; n<numtoks; n=n+2)
          {
            if (!StrCmpNR(cbuf, tokens2[n]))
            { // Match
              strcpy(cbuf1, tokens1[n]);     // Measurement
              ftmp = (float) atof(cbuf1);
              error = write(handle, &ftmp, 4);

              strcpy(cbuf1, tokens1[n+1]);   // Uncertainty
              ftmp = (float) atof(cbuf1);
              error = write(handle, &ftmp, 4);
              break;
            }

          }

        }

        nPROtot++;
        count++;

      }

    }

  }

  if (nPROtot <=0)
  {
    sprintf(Cwork1, "Total number of profiles data records = %ld", nPROtot); // TC corrected 08-19-04; changed nPRtot to nPROtot
    message(messagetitle, Cwork1);
    error = ERROR_PROFILE_SET;
    goto EXIT_ERROR;
  }

  // termination NULL
  PROcode[nPROtot][0] = '\0';

  error = 0;
  EXIT_ERROR:

  fcloseall();
  close(handle);

  return error;

}

// =============================================================================
int tokcnt(char *cp)
{
  int count, i, j, len;

  /* strip out control characters */
  len = strlen(cp);
  for(i=0; i<len; i++)
  {
    if (cp[i] < ' ')
    {
      for(j=i; j<len; j++)
      {
        cp[j] = cp[j+1];
      }
      len--;
    }
  }
  count = 0;
  if (strlen(cp) > 0 && cp[0] > ' ') count++; /* initial token */
  for(i=0; i<(int)strlen(cp); i++)
  if (cp[i] == ' ' && cp[i+1] > ' ') count++;
  return count;
}

// =============================================================================
void message(char *title, char *message)
{
    MessageBox(NULL, message, title, MB_OK);
}

// =============================================================================
void FreeTempMemFP()  // File Pointer
{
  int error;
  static size_t siz = 0;

  if (SPcodetmp != NULL)
  {
    error = getmem("SPcodetmp", (void **) &SPcodetmp, siz, __LINE__, 2);
  }

  if (PRcodetmp != NULL)
  {
    error = getmem("PRcodetmp", (void **) &PRcodetmp, siz, __LINE__, 2);
  }

  if (ADcodetmp != NULL)
  {
    error = getmem("ADcodetmp", (void **) &ADcodetmp, siz, __LINE__, 2);
  }

  if (work2 != NULL)
  {
    error = getmem("work2", (void **) &work2, siz, __LINE__, 2);
  }

  if ( pFile != NULL )
  {
     fclose( pFile );
	 pFile = NULL;
  }

// return; RFA commented out w/ 09-03-02 build
// fcloseall(); RFA commented out w/ 09-03-02 build

}

// =============================================================================
void Ljust1(char *cp, int len1) // left justify - keep trailing blanks
{
  int n, len;

  Ljust(cp);

  len = (int) strlen(cp);

  for(n=len; n<len1; n++)
  {
    cp[n] = ' ';
  }

  cp[len1] = '\0';

}

// =============================================================================
void Ljust(char *cp)
{
  int i, n, len;

  len = strlen(cp);
  for(n=0; n<len; n++)
  {
    if (!isspace(cp[n]) || cp[n] == '\n')
    {
      break;
    }
  }

  for(i=n; i<=len; i++)
  {
    cp[i-n] = cp[i];
  }

}

#define NUMPOINTERS 31
static struct zz
{
  char *ptname;
  char *pt;
  size_t size;
} ptinfo[NUMPOINTERS] =
{
  "talkarray",   NULL, 0,
  "SPcode"   ,   NULL, 0,
  "SPcode_user", NULL, 0,
  "SPcodetmp",   NULL, 0,
  "PRcode",      NULL, 0,
  "PRcode_user", NULL, 0,
  "PRcodetmp",   NULL, 0,
  "ADcode",      NULL, 0,
  "ADcodetmp",   NULL, 0,
  "SPdefsel",    NULL, 0,
  "PRdefsel",    NULL, 0,
  "PROcode",     NULL, 0,
  "tokens1",     NULL, 0,
  "tokens2",     NULL, 0,
  "tokens3",     NULL, 0,
  "Cwork",       NULL, 0,
  "Cwork1",      NULL, 0,
  "outbuf",      NULL, 0,
  "sfitbuf",     NULL, 0,
  "errbuf",      NULL, 0,
  "iwork1",      NULL, 0,
  "iSPfit",      NULL, 0,
  "jPRfit",      NULL, 0,
  "inestim",     NULL, 0,
  "CMBwork",     NULL, 0,
  "work2",       NULL, 0,
  "atotal",      NULL, 0,
  "uatotal",     NULL, 0,
  "ambSP",       NULL, 0,
  "uambSP",      NULL, 0,
  "sngval",      NULL, 0
};

// =============================================================================
int getmem(char *ptname, void **pt, size_t siz, unsigned int line, int action)
{
  char winbuf[128];
  int index;
  size_t size;

  if (action == 0)  //allocate memory
  {

    FREE(*pt);

    if (!strcmp(ptname, "talkarray"))
    {
      size = siz * sizeof(talkarray[0]) + 16;
      *pt = (char(*) [LENSTRINGS]) malloc(size);

      if (*pt == NULL)
      {
        #ifdef DEBUGNR
        sprintf(winbuf, "Memory allocation error for %s from line %lu",
        ptname, line);
        message("GETMEM", winbuf);
        exit(0);
        #else
        return ERROR_MEMORY_ALLOC;
        #endif
      }
      else
      {
        for(index=0; index<NUMPOINTERS; index++)
        {

          if (!strcmp(ptname, ptinfo[index].ptname))
          {
            ptinfo[index].size = size;
            ptinfo[index].pt = (char *) *pt;

            break;
          }

        }

        if (index < NUMPOINTERS)
        {
          memset(ptinfo[index].pt + size-16, '\252', 16);
        }
        else
        {
          sprintf(winbuf, "Pointer %s not defined in ptinfo, line = %u",
          ptname, line);
          message("GETMEM", winbuf);
          exit(0);
        }

        return 0;
      }
    }

    if (!strcmp(ptname, "SPcode")  ||
    !strcmp(ptname, "SPcode_user") ||
    !strcmp(ptname, "SPcodetmp"))
    {
      size = siz * sizeof(SPcode[0]) + 16;
      *pt = (char(*) [lenSPcode]) malloc(size);
      if (*pt == NULL)
      {
        #ifdef DEBUGNR
        sprintf(winbuf, "Memory allocation error for %s from line %lu",
        ptname, line);
        message("GETMEM", winbuf);
        exit(0);
        #else
        return ERROR_MEMORY_ALLOC;
        #endif
      }
      else
      {
        for(index=0; index<NUMPOINTERS; index++)
        {

          if (!strcmp(ptname, ptinfo[index].ptname))
          {
            ptinfo[index].size = size;
            ptinfo[index].pt = (char *) *pt;
            break;
          }

        }

        if (index < NUMPOINTERS)
        {
          memset(ptinfo[index].pt + size-16, '\252', 16);
        }
        else
        {
          sprintf(winbuf, "Pointer %s not defined in ptinfo, line = %u",
          ptname, line);
          message("GETMEM", winbuf);
          exit(0);
        }

        return 0;
      }
    }

    if (!strcmp(ptname, "PRcode")  ||
    !strcmp(ptname, "PRcode_user") ||
    !strcmp(ptname, "PRcodetmp"))
    {
      size = siz * sizeof(PRcode[0]) + 16;
      *pt = (char(*) [lenPRcode]) malloc(size);
      if (*pt == NULL)
      {
        #ifdef DEBUGNR
        sprintf(winbuf, "Memory allocation error for %s from line %lu",
        ptname, line);
        message("GETMEM", winbuf);
        exit(0);
        #else
        return ERROR_MEMORY_ALLOC;
        #endif
      }
      else
      {
        for(index=0; index<NUMPOINTERS; index++)
        {

          if (!strcmp(ptname, ptinfo[index].ptname))
          {
            ptinfo[index].size = size;
            ptinfo[index].pt = (char *) *pt;
            break;
          }

        }

        if (index < NUMPOINTERS)
        {
          memset(ptinfo[index].pt + size-16, '\252', 16);
        }
        else
        {
          sprintf(winbuf, "Pointer %s not defined in ptinfo, line = %u",
          ptname, line);
          message("GETMEM", winbuf);
          exit(0);
        }

        return 0;
      }
    }

    if (!strcmp(ptname, "ADcode") ||
    !strcmp(ptname, "ADcodetmp")
)
    {
      size = siz * sizeof(ADcode[0]) + 16;
      *pt = (char(*) [lenADcode]) malloc(size);
      if (*pt == NULL)
      {
        #ifdef DEBUGNR
        sprintf(winbuf, "Memory allocation error for %s from line %lu",
        ptname, line);
        message("GETMEM", winbuf);
        exit(0);
        #else
        return ERROR_MEMORY_ALLOC;
        #endif
      }
      else
      {
        for(index=0; index<NUMPOINTERS; index++)
        {

          if (!strcmp(ptname, ptinfo[index].ptname))
          {
            ptinfo[index].size = size;
            ptinfo[index].pt = (char *) *pt;
            break;
          }

        }

        if (index < NUMPOINTERS)
        {
          memset(ptinfo[index].pt + size-16, '\252', 16);
        }
        else
        {
          sprintf(winbuf, "Pointer %s not defined in ptinfo, line = %u",
          ptname, line);
          message("GETMEM", winbuf);
          exit(0);
        }

        return 0;
      }
    }

    if (!strcmp(ptname, "SPdefsel") ||
    !strcmp(ptname, "PRdefsel")
)
    {
      size = siz * sizeof(SPdefsel[0]) + 16;
      *pt = (char(*) [LenDefSel]) malloc(size);
      if (*pt == NULL)
      {
        #ifdef DEBUGNR
        sprintf(winbuf, "Memory allocation error for %s from line %lu",
        ptname, line);
        message("GETMEM", winbuf);
        exit(0);
        #else
        return ERROR_MEMORY_ALLOC;
        #endif
      }
      else
      {
        for(index=0; index<NUMPOINTERS; index++)
        {

          if (!strcmp(ptname, ptinfo[index].ptname))
          {
            ptinfo[index].size = size;
            ptinfo[index].pt = (char *) *pt;
            break;
          }

        }

        if (index < NUMPOINTERS)
        {
          memset(ptinfo[index].pt + size-16, '\252', 16);
        }
        else
        {
          sprintf(winbuf, "Pointer %s not defined in ptinfo, line = %u",
          ptname, line);
          message("GETMEM", winbuf);
          exit(0);
        }

        return 0;
      }
    }

    if (!strcmp(ptname, "PROcode"))
    {
      size = siz * sizeof(PROcode[0]) + 16;
      *pt = (char(*) [lenPROcode]) malloc(size);
      if (*pt == NULL)
      {
        #ifdef DEBUGNR
        sprintf(winbuf, "Memory allocation error for %s from line %lu",
        ptname, line);
        message("GETMEM", winbuf);
        exit(0);
        #else
        return ERROR_MEMORY_ALLOC;
        #endif
      }
      else
      {
        for(index=0; index<NUMPOINTERS; index++)
        {

          if (!strcmp(ptname, ptinfo[index].ptname))
          {
            ptinfo[index].size = size;
            ptinfo[index].pt = (char *) *pt;
            break;
          }

        }

        if (index < NUMPOINTERS)
        {
          memset(ptinfo[index].pt + size-16, '\252', 16);
        }
        else
        {
          sprintf(winbuf, "Pointer %s not defined in ptinfo, line = %u",
          ptname, line);
          message("GETMEM", winbuf);
          exit(0);
        }

        return 0;
      }
    }

    if (!strcmp(ptname, "tokens1") ||
    !strcmp(ptname, "tokens2") ||
    !strcmp(ptname, "tokens3")
)
    {
      size = siz * sizeof(tokens1[0]) + 16;
      *pt = (char(*) [LENTOKEN]) malloc(size);
      if (*pt == NULL)
      {
        #ifdef DEBUGNR
        sprintf(winbuf, "Memory allocation error for %s from line %lu",
        ptname, line);
        message("GETMEM", winbuf);
        exit(0);
        #else
        return ERROR_MEMORY_ALLOC;
        #endif
      }
      else
      {
        for(index=0; index<NUMPOINTERS; index++)
        {

          if (!strcmp(ptname, ptinfo[index].ptname))
          {
            ptinfo[index].size = size;
            ptinfo[index].pt = (char *) *pt;
            break;
          }

        }

        if (index < NUMPOINTERS)
        {
          memset(ptinfo[index].pt + size-16, '\252', 16);
        }
        else
        {
          sprintf(winbuf, "Pointer %s not defined in ptinfo, line = %u",
          ptname, line);
          message("GETMEM", winbuf);
          exit(0);
        }

        return 0;
      }
    }

    if (!strcmp(ptname, "Cwork") ||
    !strcmp(ptname, "Cwork1") ||
    !strcmp(ptname, "outbuf") ||
    !strcmp(ptname, "sfitbuf") ||
    !strcmp(ptname, "errbuf")
)
    {
      size = siz + 16;
      *pt = (char *) malloc(size);
      if (*pt == NULL)
      {
        #ifdef DEBUGNR
        sprintf(winbuf, "Memory allocation error for %s from line %lu",
        ptname, line);
        message("GETMEM", winbuf);
        exit(0);
        #else
        return ERROR_MEMORY_ALLOC;
        #endif
      }
      else
      {
        for(index=0; index<NUMPOINTERS; index++)
        {

          if (!strcmp(ptname, ptinfo[index].ptname))
          {
            ptinfo[index].size = size;
            ptinfo[index].pt = (char *) *pt;
            break;
          }

        }

        if (index < NUMPOINTERS)
        {
          memset(ptinfo[index].pt + size-16, '\252', 16);
        }
        else
        {
          sprintf(winbuf, "Pointer %s not defined in ptinfo, line = %u",
          ptname, line);
          message("GETMEM", winbuf);
          exit(0);
        }

        return 0;
      }
    }

    if (
    !strcmp(ptname, "iwork1") ||
    !strcmp(ptname, "iSPfit") ||
    !strcmp(ptname, "jPRfit") ||
    !strcmp(ptname, "inestim")
)
    {
      size = siz * sizeof(iwork1[0]) + 16;
      *pt = (int *) malloc(size);
      if (*pt == NULL)
      {
        #ifdef DEBUGNR
        sprintf(winbuf, "Memory allocation error for %s from line %lu",
        ptname, line);
        message("GETMEM", winbuf);
        exit(0);
        #else
        return ERROR_MEMORY_ALLOC;
        #endif
      }
      else
      {
        for(index=0; index<NUMPOINTERS; index++)
        {

          if (!strcmp(ptname, ptinfo[index].ptname))
          {
            ptinfo[index].size = size;
            ptinfo[index].pt = (char *) *pt;
            break;
          }

        }

        if (index < NUMPOINTERS)
        {
          memset(ptinfo[index].pt + size-16, '\252', 16);
        }
        else
        {
          sprintf(winbuf, "Pointer %s not defined in ptinfo, line = %u",
          ptname, line);
          message("GETMEM", winbuf);
          exit(0);
        }

        return 0;
      }
    }

    if (
    !strcmp(ptname, "CMBwork") ||
    !strcmp(ptname, "work2") ||
    !strcmp(ptname, "atotal") ||
    !strcmp(ptname, "uatotal") ||
    !strcmp(ptname, "ambSP") ||
    !strcmp(ptname, "uambSP") ||
    !strcmp(ptname, "sngval")
)
    {
      size = siz * sizeof(CMBwork[0]) + 16;
      *pt = (float *) malloc(size);
      if (*pt == NULL)
      {
        #ifdef DEBUGNR
        sprintf(winbuf, "Memory allocation error for %s from line %lu",
        ptname, line);
        message("GETMEM", winbuf);
        exit(0);
        #else
        return ERROR_MEMORY_ALLOC;
        #endif
      }
      else
      {
        for(index=0; index<NUMPOINTERS; index++)
        {

          if (!strcmp(ptname, ptinfo[index].ptname))
          {
            ptinfo[index].size = size;
            ptinfo[index].pt = (char *) *pt;
            break;
          }

        }

        if (index < NUMPOINTERS)
        {
          memset(ptinfo[index].pt + size-16, '\252', 16);
        }
        else
        {
          sprintf(winbuf, "Pointer %s not defined in ptinfo, line = %u",
          ptname, line);
          message("GETMEM", winbuf);
          exit(0);
        }

        return 0;
      }
    }


    sprintf(winbuf, "Unknown pointer %s from line %lu",
    ptname, line);
    message("GETMEM", winbuf);
    exit(0);

  }
  else if (action == 1)  //memory check
  {
    for(index=0; index<NUMPOINTERS; index++)
    {

      if (ptinfo[index].size > 0)
      {

        if (strncmp(ptinfo[index].pt+ptinfo[index].size-16,
        "\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252\252", 16))
        {
          sprintf(winbuf, "Memory overrun for pointer %s from line %lu; size = %lu",
          // rfa 2002/07/26 -- changed size to ptinfo[index].size after getting
          // warning about size being UNinitialized.
          ptinfo[index].ptname, line, ptinfo[index].size);
          message("GETMEM", winbuf);
        }

      }

    }
  }
  else if (action == 2)  //free individual memory
  {
    for(index=0; index<NUMPOINTERS; index++)
    {
      if (ptinfo[index].pt == (char *) *pt)
      {
        FREE(*pt);
        ptinfo[index].pt = NULL;
      }
    }
  }
  else if (action == 3)  //free all memory
  {
    for(index=0; index<NUMPOINTERS; index++)
    {
      FREE(ptinfo[index].pt);
    }
  }
  else
  {
    sprintf(winbuf, "Unknown action %d from line %lu",
    action, line);
    message("GETMEM", winbuf);
    exit(0);

  }

  return 0;
}

// =============================================================================
int SFitHead_c(int *Lp_type)
{
  static FILE *fp_SFit = NULL;

  int act;

  fp_SFit = fopen(SFitName, "w");

  if (fp_SFit == NULL)
  {
    sprintf(winbuf, "Not able to open output file: %s\n", SFitName);
    message1(winbuf);
    return ERROR_FILE_OPEN;
  }

  act = 1;
  SFITHEAD(&act, Lp_type, ftn_char_PRcode, ftn_char_SFit);
  SFitBuf[SFitRecLen-1] = '\0';
  fprintf(fp_SFit, "%s\n", SFitBuf);

  /*   act = 2;
  SFITHEAD(&act, Lp_type, ftn_char_PRcode, ftn_char_sfit);
  sfitbuf[sfitlen-1] = '\0';
  fprintf(fp_sfit, "%s\n", sfitbuf);

  act = 3;
  SFITHEAD(&act, Lp_type, ftn_char_PRcode, ftn_char_sfit);
  sfitbuf[sfitlen-1] = '\0';
  fprintf(fp_sfit, "%s\n", sfitbuf);

  act = 4;
  SFITHEAD(&act, Lp_type, ftn_char_PRcode, ftn_char_sfit);
  sfitbuf[sfitlen-1] = '\0';
  fprintf(fp_sfit, "%s\n", sfitbuf);
  */
  fclose(fp_SFit);
  return 0;
}

// =============================================================================
int SFitOut_c(int *Lp_type)
{
  static FILE *fp_SFit = NULL;
  int L, error;
  size_t siz;

  /*
  Lp_type = 1 => TXT
  Lp_type = 2 => CSV
  */

  /* write header */
  if (Write2DT2status)
  {
    SFitHead_c(Lp_type);
  }

  fp_SFit = fopen(SFitName, "a");

  if (fp_SFit == NULL)
  {
    sprintf(winbuf, "Not able to open output file: %s\n", SFitName);
    message1(winbuf);
    return ERROR_FILE_OPEN;
  }


  siz = 4 * nPRtot;

  error = getmem("work2", (void **) &work2, siz, __LINE__, 0);
  if (error < 0)
  {
    return error;
  }
 

  for(L=1; L<= (int) nSPtot; L++)
  {
    error = 99;
    SFITOUT(&L, Lp_type, work2, work2 + nPRtot,
    atotal, uatotal, &nSPtot, iwork1,
    ambSP, uambSP, jPRfit,
    ftn_char_SPcode, ftn_char_PRcode, ftn_char_PROcode,
    work2 + 2*nPRtot, ftn_char_SFit, &error);
    SFitBuf[SFitRecLen-1] = '\0';
    fprintf(fp_SFit, "%s\n", SFitBuf);
  }

  fclose(fp_SFit);

  return error;
}

// =============================================================================
int SFitDBF_c()
{
  int L, error;
  size_t siz;

  XBASE_FIELDS *fields = NULL;
  static int numfields;
  static int reclen;
  char *dbfbuf = NULL;
  unsigned int numrecs;
  int handle;
  int i, n;
  int type1;

  /* write header */
  if (Write2DT2status)
  {
    numfields = 13 + 2 * nPRtot;
    fields = (XBASE_FIELDS *) malloc(numfields * sizeof(fields[0]));
    if (fields == NULL)
    {
      error = ERROR_MEMORY_ALLOC;
      goto exit_error;
    }

    n = 0;
    strcpy(fields[n].name, "ELCODE");
    fields[n].type = 'C';
    fields[n].len = 6;
    fields[n].decpl = 0;

    n++;
    strcpy(fields[n].name, "ELNAME");
    fields[n].type = 'C';
    fields[n].len = 8;
    fields[n].decpl = 0;

    n++;
    strcpy(fields[n].name, "I");
    fields[n].type = 'C';
    fields[n].len = 1;
    fields[n].decpl = 0;

    n++;
    strcpy(fields[n].name, "SITEID");
    fields[n].type = 'C';
    fields[n].len = 12;
    fields[n].decpl = 0;

    n++;
    strcpy(fields[n].name, "DATE");
    fields[n].type = 'C';
    fields[n].len = 8;
    fields[n].decpl = 0;

    n++;
    strcpy(fields[n].name, "ST");
    fields[n].type = 'C';
    fields[n].len = 2;
    fields[n].decpl = 0;

    n++;
    strcpy(fields[n].name, "DR");
    fields[n].type = 'C';
    fields[n].len = 2;
    fields[n].decpl = 0;

    n++;
    strcpy(fields[n].name, "SIZE");
    fields[n].type = 'C';
    fields[n].len = 5;	      // 6!
    fields[n].decpl = 0;

    n++;
    strcpy(fields[n].name, "CONC");
    fields[n].type = 'N';
    fields[n].len = 9;
    fields[n].decpl = COMMON_CMB.DisplayDecPl;

    n++;
    strcpy(fields[n].name, "UCONC");
    fields[n].type = 'N';
    fields[n].len = 9;
    fields[n].decpl = COMMON_CMB.DisplayDecPl;

    n++;
    strcpy(fields[n].name, "RSQUAR");
    fields[n].type = 'N';
    fields[n].len = 9;
    fields[n].decpl = COMMON_CMB.DisplayDecPl;

    n++;
    strcpy(fields[n].name, "CHISQUAR");
    fields[n].type = 'N';
    fields[n].len = 9;
    fields[n].decpl = COMMON_CMB.DisplayDecPl;

    n++;
    strcpy(fields[n].name, "PCMASS");
    fields[n].type = 'N';
    fields[n].len = 9;
    fields[n].decpl = COMMON_CMB.DisplayDecPl;

    for(i=0; i<nPRtot; i++)
    {

      n++;
      strncpy(fields[n].name, PRcode[i], lenPRcode1);
      fields[n].name[lenPRcode1] = '\0';
      fields[n].type = 'N';
      fields[n].len = 9;
      fields[n].decpl = COMMON_CMB.DisplayDecPl;

      n++;
      strncpy(fields[n].name, PRcode[i]+ lenPRcode1+1, lenPRcode2);
      fields[n].name[lenPRcode2] = '\0';
      fields[n].type = 'N';
      fields[n].len = 9;
      fields[n].decpl = COMMON_CMB.DisplayDecPl;

    }

    handle = open(SFitName, O_CREAT | O_BINARY | O_RDWR, _S_IREAD | _S_IWRITE);
    if (handle == -1)
    {
      error = ERROR_FILE_OPEN;
      goto exit_error;
    }

    error = Create(handle, fields, numfields, &dbfbuf, &reclen);
    if (error)
    {
      error = ERROR_FILE_OPEN;
      goto exit_error;
    }

    error = close(handle);
    if (error)
    {
      error = ERROR_FILE_CLOSE;
    }

    FREE(fields);
    FREE(dbfbuf);

  }

  handle = open(SFitName, O_BINARY | O_RDWR);
  if (handle == -1)
  {
    error = ERROR_FILE_OPEN;
    goto exit_error;
  }

  error = loadDBinfo(handle, &fields, &numfields, &numrecs,
  &reclen, &dbfbuf);
  if (error)
  {
    error = ERROR_FILE_OPEN;
    goto exit_error;
  }

  siz = 4 * nPRtot;

  error = getmem("work2", (void **) &work2, siz, __LINE__, 0);
  if (error < 0)
  {
    return error;
  }


  type1 = 3;

  for(L=1; L<= (int) nSPtot; L++)
  {
    error = 99;
    SFITOUT(&L, &type1, work2, work2 + nPRtot,
    atotal, uatotal, &nSPtot, iwork1,
    ambSP, uambSP, jPRfit,
    ftn_char_SPcode, ftn_char_PRcode, ftn_char_PROcode,
    work2 + 2*nPRtot, ftn_char_SFit, &error);

    error = AppendRec(handle, SFitBuf);
    if (error)
    {
      error = ERROR_FILE_WRITE;
      goto exit_error;
    }
  }


  error = 0;
  exit_error:

  error = close(handle);
  if (error)
  {
    error = ERROR_FILE_CLOSE;
  }

  FREE(fields);
  FREE(dbfbuf);

  return error;
}

// =============================================================================
int SFitWKS_c()
{
  char *seps = " \n";
  char *cp;
  static unsigned short row, col;
  int act;
  int type;
  size_t siz;
  int L, Lerror;
  double dtmp;
  int error;

  /*
  Lp_type = 4 => WKS
  */

  /* write WKS header and data header */
  if (Write2DT2status)
  {
    error = writeWKSheader();
    if (error)
    {
      return error;
    }

    act = 1;
    type = 1;

    SFITHEAD(&act, &type, ftn_char_PRcode, ftn_char_SFit);
    SFitBuf[SFitRecLen-1] = '\0';

    row = 0;
    col = 0;
    cp = strtok(SFitBuf, seps);
    while(cp)
    {
      error = WKSlabel(row, col, cp);
      if (error)
      {
        return error;
      }
      cp = strtok(NULL, seps);
      col++;
    }

  }

  /* write data */

  siz = 4 * nPRtot;
  error = getmem("work2", (void **) &work2, siz, __LINE__, 0);
  if (error < 0)
  {
    return error;
  }


  for(L=1; L<= (int) nSPtot; L++)
  {
    Lerror = 99;
    type = 1;
    SFITOUT(&L, &type, work2, work2 + nPRtot,
    atotal, uatotal, &nSPtot, iwork1,
    ambSP, uambSP, jPRfit,
    ftn_char_SPcode, ftn_char_PRcode, ftn_char_PROcode,
    work2 + 2*nPRtot, ftn_char_SFit, &Lerror);
    SFitBuf[SFitRecLen-1] = '\0';
    if (Lerror)
    {
      return(int) error;
    }


    row++;
    col = 0;

    cp = strtok(SFitBuf, seps);
    while(cp)
    {
      if (
      col == 0 || /* element code */
      col == 1 || /* element name */
      col == 2 || /* selected code */
      col == 3 || /* site id */
      col == 4 || /* date */
      col == 7    /* size */
)
      {
        error = WKSlabel(row, col, cp);
        if (error)
        {
          return error;
        }
      }
      else
      {
        dtmp = atof(cp);
        error = WKSreal(row, col, dtmp);
        if (error)
        {
          return error;
        }
      }

      cp = strtok(NULL, seps);
      col++;
    }

  }

  return 0;
}

// =============================================================================
int AppendRec(int handle, char *dbfbuf)
{
  /*
  appends record buffer for record rec in dbfbuf.
  NB rec = 0 for first record
  */

  int error;
  unsigned int headerlen;
  unsigned int reclen;
  unsigned int numrecs;
  unsigned int pos;

  XBASE_HEADER head;

  error = lseek(handle, 0, SEEK_SET);
  if (error < 0) return ERROR_FILE_WRITE;

  error = read(handle, (char *) &head, 32);
  if (error <= 0) return ERROR_FILE_WRITE;

  numrecs = head.numrecs3;
  numrecs = 256 * numrecs + head.numrecs2;
  numrecs = 256 * numrecs + head.numrecs1;
  numrecs = 256 * numrecs + head.numrecs0;
  headerlen = head.headlen1;
  headerlen = 256 * headerlen + head.headlen0;
  reclen = head.reclen1;
  reclen = 256 * reclen + head.reclen0;

  pos = headerlen + numrecs * reclen;

  error = lseek(handle, pos, SEEK_SET);
  if (error < 0) return ERROR_FILE_WRITE;

  dbfbuf[0] = ' '; /* nondelete flag */

  error = write(handle, dbfbuf, reclen);
  if (error < 0) return ERROR_FILE_WRITE;

  /* update header */
  numrecs++;
  head.numrecs0 = (unsigned char) numrecs % 256;
  numrecs = (numrecs - head.numrecs0) / 256;
  head.numrecs1 = (unsigned char) numrecs % 256;
  numrecs = (numrecs - head.numrecs1) / 256;
  head.numrecs2 = (unsigned char) numrecs % 256;
  head.numrecs3 = (unsigned char)(numrecs - head.numrecs2) / 256;

  error = lseek(handle, 0L, SEEK_SET);
  if (error < 0) return ERROR_FILE_WRITE;

  error = write(handle, &head, 32);
  if (error < 0) return ERROR_FILE_WRITE;

  return 0;

}

// =============================================================================
int loadDBinfo(int handle, XBASE_FIELDS **fields, int *p_numfields,
unsigned int *p_numrecs, int *p_reclen, char **dbfbuf)
{

  XBASE_HEADER header;
  int numrecs;
  int headerlen;
  int reclen;
  int fld, offset, error, numfields;
  unsigned char cbuf32[32];
  int Lerror;

  Lerror = lseek(handle, 0L, SEEK_SET); /* BOF */
  if (Lerror == -1) return ERROR_FILE_READ;  /* EXIT */

  error = read(handle, &header, 32);
  if (error != 32) return ERROR_FILE_READ;  /* EXIT */

  numrecs = header.numrecs3;
  numrecs = 256 * numrecs + header.numrecs2;
  numrecs = 256 * numrecs + header.numrecs1;
  numrecs = 256 * numrecs + header.numrecs0;
  headerlen = header.headlen1;
  headerlen = 256 * headerlen + header.headlen0;
  reclen = header.reclen1;
  reclen = 256 * reclen + header.reclen0;

  if (*dbfbuf != NULL) return ERROR_FILE_READ;
  *dbfbuf = malloc((size_t) reclen);
  if (*dbfbuf == NULL) return ERROR_FILE_READ; /* EXIT */

  numfields = (headerlen / 32) - 1;
  if (numfields > 255) return ERROR_FILE_READ;

  if (*fields != NULL) return ERROR_FILE_READ;
  *fields = (XBASE_FIELDS *) malloc(
(size_t)(numfields * sizeof(*fields[0])));

  if (*fields == NULL) return ERROR_FILE_READ; /* EXIT */

  /* read input file fields info */
  fld = 0;
  offset = 1;
  for(fld=0; fld<numfields; fld++)
  {
    error = read(handle, cbuf32, 32);
    if (error != 32) return 8;  /* EXIT */

    strncpy((*fields) [fld].name, cbuf32, 10);
    (*fields) [fld].name [10] = '\0';
    (*fields) [fld].type = *(cbuf32 + 11);
    (*fields) [fld].len = (int)(*(cbuf32 + 16));
    (*fields) [fld].decpl = (int)(*(cbuf32 + 17));
    (*fields) [fld].offset = offset;
     offset += (*fields) [fld].len;
  }

  *p_numfields = numfields;
  *p_numrecs = numrecs;
  *p_reclen = reclen;

  return 0;

}

// =============================================================================
int Create(int handle, XBASE_FIELDS *fields, int numfields, char **dbfbuf, int *reclen)
{
  XBASE_HEADER header;
  unsigned char cbuf32[32];
  int year, month, day;
  unsigned int headerlen;
  int error, i, fld, offset;
  struct stat statbuf;

  error = fstat(handle, &statbuf);
  if (error == -1)
  {
    return ERROR_FILE_WRITE;
  }

  #ifdef UNIX
  error = ftruncate(handle, 0L);
  #else
  error = chsize(handle, 0L);
  #endif
  if (error == -1)
  {
    return ERROR_FILE_WRITE;
  }

  if (fields == NULL)
  {
    return ERROR_FILE_WRITE;
  }

  /* determine reclen and set offset */
  offset = 1;
  for(fld=0; fld<numfields; fld++)
  {
    fields[fld].offset = offset;
    offset += fields[fld].len;
  }

  *reclen = offset;

  /* allocate memory for record buffer */
  if (*dbfbuf != NULL) return 4; /* EXIT */

  *dbfbuf = malloc((size_t) *reclen);
  if (*dbfbuf == NULL) return 5; /* EXIT */

  /* write header */
  memset(&header, '\000', 32);
  NRdate(&year, &month, &day);
  header.version = '\003'; /* no .dbt */
  header.yy = (char) year;
  header.mm = (char) month;
  header.dd = (char) day;
  /* numrecs = 0 */
  headerlen = 32 *(numfields + 1) + 1;
  header.headlen0 = (unsigned char)(headerlen % 256); /* Least Significant Bytes */
  header.headlen1 = (unsigned char)(headerlen / 256);
  header.reclen0 = (unsigned char)(*reclen % 256);    /* Least Significant Bytes */
  header.reclen1 = (unsigned char)(*reclen / 256);

  error = write(handle, &header, 32);
  if (error == -1) return ERROR_FILE_WRITE; /* EXIT */


  for(i=0; i<numfields; i++)
  {
    memset(&cbuf32, '\000', 32);
    strncpy(cbuf32, fields[i].name, 10);
    *(cbuf32 + 11) = fields[i].type;
    *(cbuf32 + 16) = (unsigned char) fields[i].len;
    *(cbuf32 + 17) = (unsigned char) fields[i].decpl;
    error = write(handle, cbuf32, 32);
    if (error == -1) return ERROR_FILE_WRITE; /* EXIT */
  }

  error = write(handle, "\015", 1); /* CR */
  if (error == -1) return ERROR_FILE_WRITE; /* EXIT */

  return 0;

}

// =============================================================================
void NRdate(int *year, int *month, int *day)
/* returns data as YY, MM, DD */
{
  struct tm *tim;
  time_t aclock;

  time(&aclock);
  tim = localtime(&aclock);
  *year = tim->tm_year;
  *month = tim->tm_mon + 1;
  *day = tim->tm_mday;

}

// =============================================================================
int dbf2txt(char *infile, char *outfile)
{
  XBASE_HEADER head; /* first 32 bytes of header */

  XBASE_FIELDS *fields = NULL;

  char cbuf100[100];

  int error, i, nfld;
  int rec, numrecsin;
  unsigned int pos;
  unsigned int headerlen;
  int reclen;
  int in_handle;
  int numfields, offset;

  char *dbfbuf_in = NULL;
  static FILE *fp_out = NULL;


  in_handle = open(infile, O_BINARY | O_RDONLY);

  if (in_handle == -1)
  {
    error = ERROR_FILE_OPEN;
    goto exit_error;
  }

  fp_out = fopen(outfile, "w");
  if (fp_out == NULL)
  {
    error = ERROR_FILE_OPEN;
    goto exit_error;
  }

  /* read header info */

  error = read(in_handle, (char *) &head, 32);
  if (error <= 0)
  {
    error = ERROR_FILE_READ;
    goto exit_error;
  }
  numrecsin = head.numrecs3;
  numrecsin = 256 * numrecsin + head.numrecs2;
  numrecsin = 256 * numrecsin + head.numrecs1;
  numrecsin = 256 * numrecsin + head.numrecs0;
  headerlen = head.headlen1;
  headerlen = 256 * headerlen + head.headlen0;
  reclen = head.reclen1;
  reclen = 256 * reclen + head.reclen0;

  dbfbuf_in = malloc((size_t) reclen);
  if (dbfbuf_in == NULL)
  {
    error = ERROR_MEMORY_ALLOC;
    goto exit_error;
  }

  numfields = (headerlen / 32) - 1;
  fields = (XBASE_FIELDS *) malloc(numfields * sizeof(fields[0]));
  if (fields == NULL)
  {
    error = ERROR_MEMORY_ALLOC;
    goto exit_error;
  }


  /* read fields info */
  i = 0;
  offset = 1;
  pos = 32;
  for(nfld=0; nfld<numfields; nfld++)
  {
    error = read(in_handle, cbuf100, 32);
    if (error <= 0)
    {
      error = ERROR_FILE_READ;
      goto exit_error;
    }
    strncpy(fields [nfld].name, cbuf100, 10);
    fields [nfld].name [10] = '\0';
    fields [nfld].type = *(cbuf100 + 11);
    fields [nfld].len = (int) *(cbuf100 + 16);
    fields [nfld].decpl = (int) *(cbuf100 + 17);
    fields [nfld].offset = offset;
    offset += fields [nfld].len;
    pos += 32;
  }

  /* read to end of header */
  error = read(in_handle, cbuf100, headerlen - pos);
  if (error <= 0)
  {
    error = ERROR_FILE_READ;
    goto exit_error;
  }


  /* print header */
  for(i=0; i<numfields; i++)
  {
    error = fprintf(fp_out, "%s ", fields[i].name);
    if (error < 0)
    {
      error = ERROR_FILE_WRITE;
      goto exit_error;
    }
  }
  error = fprintf(fp_out, "\n");
  if (error < 0)
  {
    error = ERROR_FILE_WRITE;
    goto exit_error;
  }

  /* process records */
  for(rec=0; rec<numrecsin; rec++)
  {
    error = read(in_handle, dbfbuf_in, reclen);
    if (error <= 0)
    {
      error = ERROR_FILE_READ;
      goto exit_error;
    }
    pos = 1;
    for(i=0; i<numfields; i++)
    {
      strncpy(cbuf100, dbfbuf_in+pos, fields[i].len);
      cbuf100[fields[i].len] = '\0';

      if (fields[i].type == 'D')
      {
        cbuf100[0] = cbuf100[4];  // output MM/DD/YY date format
        cbuf100[1] = cbuf100[5];
        cbuf100[4] = cbuf100[6];
        cbuf100[5] = cbuf100[7];
        cbuf100[6] = cbuf100[2];
        cbuf100[7] = cbuf100[3];
        cbuf100[2] = '/';
        cbuf100[3] = cbuf100[4];
        cbuf100[4] = cbuf100[5];
        cbuf100[5] = '/';
      }

      error = fprintf(fp_out, "%s ", cbuf100);
      if (error < 0)
      {
        error = ERROR_FILE_READ;
        goto exit_error;
      }
      pos += fields[i].len;
    }
    error = fprintf(fp_out, "\n");
    if (error < 0)
    {
      error = ERROR_FILE_READ;
      goto exit_error;
    }
  }

  error = 0;
  exit_error:

  FREE(dbfbuf_in);
  FREE(fields);

  close(in_handle);
  fclose(fp_out);
  return error;

}

// =============================================================================
int wks2txt(char *infile, char *outfile)
{

  static FILE *fp_out = NULL;

  int handle_in;
  unsigned opcode, reclen;
  int error, error1, error5;
  unsigned char cbuf[10000];
  unsigned short *p_ushort;
  short *p_short;
  unsigned short  row, row1, col;
  short  intv;
  double *p_d, dtmp;
  unsigned short numrows, numcols;

  handle_in = open(infile, O_BINARY | O_RDONLY);
  if (handle_in == -1)
  {
    error = ERROR_FILE_OPEN;
    goto exit_error;
  }

  fp_out = fopen(outfile, "w");
  if (fp_out == NULL)
  {
    error = ERROR_FILE_OPEN;
    goto exit_error;
  }

  p_ushort = (unsigned short *)(cbuf+4);
  p_short = (short *)(cbuf+4);
  p_d = (double *)(cbuf+8);

  numrows = 0;
  numcols = 0;
  while(1) /* count rows and columns */
  {

    error1 = read(handle_in, cbuf, 4);
    if (error1 == 0 || error1 == -1)
    {
      break;
    }

    p_ushort = (unsigned short *)(cbuf);
    opcode = p_ushort[0];
    reclen = p_ushort[1];

    error5 = read(handle_in, cbuf+4, reclen);

    p_ushort = (unsigned short *)(cbuf+5);
    col = p_ushort[0];
    row = p_ushort[1];

    if (opcode == 0x0f ||  /* label */
    opcode == 0x0d ||  /* integer */
    opcode == 0x0e)  /* double */
    {
      numrows = max(row, numrows);
      numcols = max(col, numcols);
    }

  }

  numrows++;
  numcols++;

  for(row=0; row<numrows; row++)
  {

    lseek(handle_in, 0L, SEEK_SET); /* rewind */

    while(1)
    {
      error1 = read(handle_in, cbuf, 4);
      if (error1 == 0 || error1 == -1)
      {
        break;
      }

      p_ushort = (unsigned short *)(cbuf);
      opcode = p_ushort[0];
      reclen = p_ushort[1];

      error5 = read(handle_in, cbuf+4, reclen);

      p_ushort = (unsigned short *)(cbuf+5);
      col = p_ushort[0];
      row1 = p_ushort[1];

      if (row == row1)
      {
        if (opcode == 0x0f)  /* label */
        {
          fprintf(fp_out, "%s ", cbuf+10); /* skip initial ' */
        }

        else if (opcode == 0x0d)  /* 2byte integer */
        {
          p_short = (short *)(cbuf+5);
          intv = p_short[2];
          fprintf(fp_out, "%d ", intv);
        }

        else if (opcode == 0x0e)  /* double */
        {
          p_d = (double *)(cbuf+9);
          dtmp = p_d[0];
          fprintf(fp_out, "%f ", dtmp);
        }

      }

    }

    fprintf(fp_out, "\n", row);

  }


  error = 0;
  exit_error:

  close(handle_in);
  fclose(fp_out);

  return error;

}

// =============================================================================
int car2txt(char *infile, char *outfile)
{
  char *seps = " \n";
  unsigned int numlines, numtoks, lentoks, pos; // linSPength
  char *tok;
  int nline, ntok;
  char *cline;
  char *p_tok;
  char *tokarray;
  static FILE *fp_in = NULL;
  static FILE *fp_out = NULL;

  unsigned int dimline, dimline1;

  fp_in  = fopen(infile, "r");
  if (fp_in == NULL)
  {
    return ERROR_FILE_OPEN;
  }

  fp_out = fopen(outfile, "w");
  if (fp_out == NULL)
  {
    return ERROR_FILE_OPEN;
  }

  /* read through input file */

  numlines = 0;
  numtoks  = 0;
  lentoks = 0;

  dimline = 16000;
  cline = malloc((size_t) dimline);
  if (cline == NULL)
  {
    return ERROR_MEMORY_ALLOC;
  }

  dimline1 = 0;
  while((p_tok = fgets(cline, dimline, fp_in)))
  {
    dimline1 = max(dimline1, strlen(cline));
    p_tok = strtok(cline, seps);
    if (p_tok)
    {
      numlines++;
      if (numlines == 1)
      {
        numtoks++;
      }
      lentoks = max(lentoks, strlen(p_tok));

      while(p_tok = strtok(NULL, seps))
      {
        if (numlines == 1)
        {
          numtoks++;
        }
        lentoks = max(lentoks, strlen(p_tok));
      }

    }

  }

  /*
  allocate array space
  */
  cline = (char *) realloc(cline, (size_t)(dimline1+1));
  if (cline == NULL)
  {
    return ERROR_MEMORY_ALLOC;
  }

  tok = malloc((size_t)(lentoks + 1));
  if (tok == NULL)
  {
    FREE(cline);
    return ERROR_MEMORY_ALLOC;
  }

  tokarray = malloc((size_t)(numtoks * numlines * lentoks));
  if (tokarray == NULL)
  {
    FREE(cline);
    FREE(tok);
    return ERROR_MEMORY_ALLOC;
  }

  memset(tokarray, ' ', numtoks * numlines * lentoks);

  rewind(fp_in);

  for(nline=0; nline<numlines; nline++)
  {
    fgets(cline, dimline, fp_in);
    p_tok = strtok(cline, seps);
    pos = nline * numtoks * lentoks;
    strncpy(tokarray+pos, p_tok, lentoks);
    for(ntok=1; ntok<numtoks; ntok++)
    {
      p_tok = strtok(NULL, seps);
      if (!p_tok)
      {
        break;
      }
      pos = pos + lentoks;
      strncpy(tokarray+pos, p_tok, lentoks);
    }

  }


  for(ntok=0; ntok<numtoks; ntok++)
  {
    for(nline=0; nline<numlines; nline++)
    {
      pos = (nline * numtoks + ntok) * lentoks;
      strncpy(cline, tokarray+pos, lentoks);
      cline[lentoks] = '\0';
      fprintf(fp_out, "%s ", cline);
    }
    fprintf(fp_out, "\n");
  }

  fclose(fp_in);
  fclose(fp_out);
  FREE(cline);
  FREE(tok);
  FREE(tokarray);

  return 0;

}

// =============================================================================
int csv2txt(char *infile, char *outfile)
{

  #define DIMLINE 10000
  static char line1[DIMLINE];

  static FILE *fp_in = NULL;
  static FILE *fp_out = NULL;

  #define DIMBUF 100
  char *cp, *Cerror;
  int numtoks, n, pos;

  fp_in = fopen(infile, "r");
  if (fp_in == NULL)
  {
    return ERROR_FILE_OPEN;
  }

  fp_out = fopen(outfile, "w");
  if (fp_out == NULL)
  {
    return ERROR_FILE_OPEN;
  }

  // first line
  numtoks = 0;

  Cerror = fgets(line1, DIMLINE, fp_in);
  line1[ strlen(line1) - 1 ] = '\0';  // line feed
  if (Cerror == NULL) return ERROR_FILE_READ;

  pos = 0;
  while(1)
  {
    cp = StrTokNR1(line1, &pos);
    if (strcmp(cp, ""))
    {
      fprintf(fp_out, "%s ", cp);
      numtoks++;
    }
    else
    {
      break;
    }
  }
  fprintf(fp_out, "\n");

  while(1)
  {
    Cerror = fgets(line1, DIMLINE, fp_in);
    line1[ strlen(line1) - 1 ] = '\0';  // line feed
    if (Cerror == NULL) break;
    pos = 0;
    for(n=0; n<numtoks; n++)
    {
      cp = StrTokNR1(line1, &pos);
      if (strcmp(cp, ""))
      {
        fprintf(fp_out, "%s ", cp);
      }
      else
      {
        fprintf(fp_out, "%s ", "_");
      }

    }

    fprintf(fp_out, "\n");

  }

  fclose(fp_in);
  fclose(fp_out);
  return 0;
}

// =============================================================================
char *StrTokNR1(char *str, int *p_pos)
// Return tokens for CSV records; modifies str like strtok
{
  int n, pos1;

  pos1 = *p_pos;

  for(n=0; n < strlen(str + pos1); n++)
  {
    if (*(str+pos1+n) == ',')
    {
      *(str+pos1+n) = '\0';

      *p_pos += n + 1;

      // right justify
      if (n > 0)
      {
        while(*(str+pos1+n-1) <= ' ')
        {
          n--;
          *(str+pos1+n) = '\0';
        }

      }

      // left justify
      if (n > 0)
      {
        while(*(str+pos1) <= ' ')
        {
          pos1++;
        }
      }

      return str + pos1;
    }
  }

  *p_pos = strlen(str);
  return str + pos1;

}

// =============================================================================
int writeWKSheader()
{

  int handle_out, error;

  handle_out = open(SFitName,  O_CREAT, S_IWRITE);
  close(handle_out);
  handle_out = open(SFitName,  O_BINARY | O_RDWR | O_TRUNC);
  if (handle_out == -1)
  {
    error = ERROR_FILE_OPEN;
    goto exit_error;
  }

  /* write WKS header */

  error = write(handle_out, wkshead, HEADLEN);
  if (error != HEADLEN)
  {
    error = ERROR_FILE_WRITE;
    goto exit_error;
  }

  error = write(handle_out, wkstail, TAILLEN);
  if (error != TAILLEN)
  {
    error = ERROR_FILE_WRITE;
    goto exit_error;
  }

  close(handle_out);

  error = 0;
  exit_error:
  return error;
}

// =============================================================================
int WKSlabel(unsigned short row, unsigned short col, char *cp)
{
  int handle_out, error;
  unsigned int recl;
  int Lerror;

  handle_out = open(SFitName,  O_BINARY | O_RDWR);
  if (handle_out == -1)
  {
    error = ERROR_FILE_OPEN;
    goto exit_error;
  }

  Lerror = lseek(handle_out, -TAILLEN, SEEK_END); /* write over tail */
  if (Lerror == -1L)
  {
    error = ERROR_FILE_WRITE;
    goto exit_error;
  }


  recl = 7 + strlen(cp); /* record length; 6+1 include terminating null */

  *((unsigned short *)(label+2)) = (unsigned short) recl;

  *((unsigned short *)(label+5)) = col;
  *((unsigned short *)(label+7)) = row;

  strcpy(label+10, cp);

  recl = recl + 4;
  error = write(handle_out, label, recl);

  if (error != recl)
  {
    error = ERROR_FILE_WRITE;
    goto exit_error;
  }

  error = write(handle_out, wkstail, TAILLEN);
  if (error != TAILLEN)
  {
    error = ERROR_FILE_WRITE;
    goto exit_error;
  }

  close(handle_out);
  error = 0;

  exit_error:
  return error;

}

// =============================================================================
int WKSreal(unsigned short row, unsigned short col, double dtmp)
{
  int handle_out, error;
  int Lerror;

  handle_out = open(SFitName,  O_BINARY | O_RDWR);
  if (handle_out == -1)
  {
    error = ERROR_FILE_OPEN;
    goto exit_error;
  }

  Lerror = lseek(handle_out, -TAILLEN, SEEK_END); /* write over tail */
  if (Lerror == -1L)
  {
    error = ERROR_FILE_WRITE;
    goto exit_error;
  }

  *((double *)(real+9)) = dtmp;

  *((unsigned short *)(real+5)) = col;
  *((unsigned short *)(real+7)) = row;

  error = write(handle_out, real, 17);
  if (error != 17)
  {
    error = ERROR_FILE_WRITE;
    goto exit_error;
  }

  error = write(handle_out, wkstail, TAILLEN);
  if (error != TAILLEN)
  {
    error = ERROR_FILE_WRITE;
    goto exit_error;
  }

  close(handle_out);
  error = 0;

  exit_error:
  return error;

}

// =============================================================================
void __stdcall MESSAGEBOXC_(FSTRING desc) // Here, MESSAGEBOXC must be ALL CAPS -tc
{
    int len = min(desc.len, 254);
    strncpy(cbuf, desc.addr, len);
    cbuf[254] = '\0';
    MessageBox(NULL, cbuf, "EPA-CMB8.2 Fortran Section", MB_OK); // TC revised 10/27/04
}

// =============================================================================
void message1(char *mess)
{
    MessageBox(NULL, mess, "CMB82.c - DEBUG", MB_OK);
}

// =============================================================================
int StuffTokArray(char *toks, int maxnum, int width,  char *cp)
{
  /*
  This procedure parses a line - cp - and places the tokens in toks;
  width gives the width of each element of toks, including terminating \0
  MaxNum gives the maximum number of tokens separators are assummed to be 
  blanks the number of tokens is returned -1 is returned on width error
  */

  unsigned len, n, n0;
  int num;

  /* eliminate redundant white space */
  SepToks(cp);
  len = strlen(cp);
  if (len == 0)
  {
    return 0;
  }

  n0 = 0;
  num = 0;
  for(n = 0; n < len; n++)
  {
    if (cp[n+1] == ' ' || cp[n+1] == '\n' || cp[n+1] == '\0')
    {
      if (n - n0 + 2 > width)
      {
        return -1;
      }
      if (num > maxnum -1)
      {
        return -2;
      }
      strncpy(toks + num*width, cp+n0, n-n0+1);
      *(toks + num*width + n-n0+1) = '\0';
      num++;
      if (num == maxnum)
      {
        return num;
      }
      n = n+1;
      n0 = n+1;
    }
  }

  return num;

}

// =============================================================================
void SepToks(char *cp)
{ /* separates tokens with one white space */
  int i, n, len;

  Ljust(cp);
  Rjust(cp);

  len = strlen(cp);
  for(i=0; i<len; i++)
  {
    if (cp[i] == ' ' && cp[i+1] == ' ')
    {
      for(n=i; n<len; n++)
      {
        cp[n] = cp[n+1];
      }
      i--;
      len--;
    }
  }

}

// =============================================================================
void ParseAmbient(char *toks, int width, char *ADcode)
{
  /*
  This procedure parses an ambient data record;
  width gives the width of each element of toks, including terminating \0
  separators are assummed to be blanks
  */
  int pos, ntok, lenAD;

  ntok = 0;
  pos = 0;
  lenAD = lenADcode1; //siteid
  strncpy(toks + ntok*width, ADcode+pos, lenAD);
  *(toks + ntok*width + lenAD) = '\0';

  ntok++;
  pos += lenAD+1; //date
  lenAD = lenADcode2;
  strncpy(toks + ntok*width, ADcode+pos, lenAD);
  *(toks + ntok*width + lenAD) = '\0';

  ntok++;
  pos += lenAD+1; //duration
  lenAD = lenADcode3;
  strncpy(toks + ntok*width, ADcode+pos, lenAD);
  *(toks + ntok*width + lenAD) = '\0';

  ntok++;
  pos += lenAD+1; //sthr
  lenAD = lenADcode4;
  strncpy(toks + ntok*width, ADcode+pos, lenAD);
  *(toks + ntok*width + lenAD) = '\0';

  ntok++;
  pos += lenAD+1; //size
  lenAD = lenADcode5;
  strncpy(toks + ntok*width, ADcode+pos, lenAD);
  *(toks + ntok*width + lenAD) = '\0';

}

// =============================================================================
void Rjust(char *cp)
{
  int i;

  for(i = strlen(cp) - 1; i >= 0; i--)
  if (*(cp + i) > ' ' &&(int)(*(cp + i)) < 127)
  {
    *(cp + i + 1) = '\0';
    break;
  }
}

// =============================================================================
int StrCmpNR(char cp1[], char cp2[]) {
  int i1, i2, n1, n2, len1, len2;

  /* this function compares two tokens ignoring leading and trailing
  nonprinting characters */

  len1 = strlen(cp1);
  len2 = strlen(cp2);

  for(i1=0; i1<len1; i1++)
  if (cp1[i1] > ' ') break;

  for(i2=0; i2<len2; i2++)
  if (cp2[i2] > ' ') break;

  for(n1=len1; n1>=0; n1--)
  if (cp1[n1] > ' ') break;

  for(n2=len2; n2>=0; n2--)
  if (cp2[n2] > ' ') break;

  if (n1 < i1 && n2 < i2) return 0; /* both blank */

  if (n1 - i1 != n2 -i2) return 1;

  return strncmp(cp1+i1, cp2+i2, n1 - i1 + 1);

}
